/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package common;
import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.StringTokenizer;

import math.VecMath;
import pattern.BaseLightPattern;
import pattern.CalculationRecords;
import pattern.ILightPattern;
import pattern.LightChar;
import pattern.LightCharPattern;


public class FileIO<T extends Number> {
	

	private final static String BASE_CONFIG_DIR = "config";
	private final static String RECORD_DIR = "records";
	private final static String CHAR_DIR = "lightpatterns/lightchars";
	private final static String PATTERN_DIR = "lightpatterns";
	private final static String MUSEUM_DIR = "museum";
	private final static String INFO_DIR = "info";
	private final static String THEME_DIR = "themes";
	private final static String CHARS_FILE = "char_map.txt";
	private final static String MUSEUM_FILE = "patterns.txt";
	private final static String TAIL = ".txt";
	private static final String CONFIG_LIST_FILE = "cfg_list.txt";
	private static String USER_CONFIG_DIR;
	private String configDir;
	private final VecMath<T> math;
	public FileIO(VecMath<T> math) {
		this.math = math;
		if(USER_CONFIG_DIR!=null)
			configDir = USER_CONFIG_DIR;
		else
			configDir = BASE_CONFIG_DIR;
	}
	public FileIO(VecMath<T> math,String ucd) {
		this.math = math;
		setUserCfgDir(ucd);
	}
	public void setUserCfgDir(String userCfgDir) {
		USER_CONFIG_DIR = userCfgDir;
		if(USER_CONFIG_DIR!=null)
			configDir = USER_CONFIG_DIR;
		else
			configDir = BASE_CONFIG_DIR;
	}
	public static String getUserCfgDir() {
		return USER_CONFIG_DIR;
	}
	public static String getUserPatternDir() {
		return USER_CONFIG_DIR+File.separator+PATTERN_DIR;
	}
	public static String getUserRecordDir() {
		return USER_CONFIG_DIR+File.separator+RECORD_DIR;
	}
	public static String getUserMuseumDir() {
		return USER_CONFIG_DIR+File.separator+MUSEUM_DIR;
	}
	public static boolean checkUserConfigDir() {
		if(getUserCfgDir()==null)
			return false;
		File dir = new File(getUserCfgDir());
		if(!dir.exists())
			return false;
		dir = new File(getUserPatternDir());
		if(!dir.exists())
			return false;
		dir = new File(getUserRecordDir());
		if(!dir.exists())
			return false;
		return true;
	}
	public static void createUserConfigDir() throws Exception{
		File dir = new File(getUserCfgDir());
		if(!dir.exists())
			dir.mkdir();
		dir = new File(getUserPatternDir());
		if(!dir.exists())
			dir.mkdir();
		dir = new File(getUserRecordDir());
		if(!dir.exists())
			dir.mkdir();
		dir = new File(getUserMuseumDir());
		if(!dir.exists())
			dir.mkdir();
	}
	public void installUserConfigs() throws Exception {
		configDir = BASE_CONFIG_DIR;
		List<BaseLightPattern<T>> patterns = loadLightPatterns();
		List<CalculationRecords<T>> records = new ArrayList<CalculationRecords<T>>();
		for(BaseLightPattern<T> blp:patterns)
			if(blp.isStoreCalcEnabled())
				records.add(loadRecords(blp.getName()));
		List<BaseLightPattern<T>> list = loadMuseum(patterns);
		configDir = USER_CONFIG_DIR;
		for(BaseLightPattern<T> blp:patterns)
			save2(blp);	
		for(CalculationRecords<T> rec:records)
			save2(rec);
		save2Museum(list);
	}
	private void save2(CalculationRecords<T> recs) throws Exception {
		File outFile = findFile(getUserRecordDir(),recs.getPatternName()+TAIL);
		if(outFile==null){
			String path = getUserRecordDir()+File.separator;
			path += recs.getPatternName()+TAIL;
			outFile = new File(path);
		}
		writeFile(outFile,recs.toString());
	}
	private void save2Museum(List<BaseLightPattern<T>> patterns) throws Exception {
		File outFile = findFile(getUserMuseumDir(),MUSEUM_FILE);
		if(outFile==null){
			String path = getUserMuseumDir()+File.separator;
			path += MUSEUM_FILE;
			outFile = new File(path);
		}
		String list = "";
		for(BaseLightPattern<T> blp:patterns)
			list += blp.getName()+"\n";
		writeFile(outFile,list.toString());
	}
	public void save(CalculationRecords<T> recs) {
		try {
			save2(recs);
		} catch(Exception e) {
			Log.getIstc().log("Error: Can't store calculation ");
			Log.getIstc().logln("record "+recs.getPatternName()+".");
			Log.getIstc().logln(e.getMessage());
		}
	}
	public void save(CalculationRecords<T> recs,File f){
		try{
			writeFile(f,recs.toString());
		} catch(Exception e) {
			Log.getIstc().log("Error: Can't store calculation ");
			Log.getIstc().logln("record to file "+f.getPath()+".");
			Log.getIstc().logln(e.getMessage());
		}		
	}
	private void save2(ILightPattern<T> lp) throws Exception {
		File outFile = findFile(getUserPatternDir(),lp.getName()+TAIL);
		if(outFile==null){
			String path = getUserPatternDir()+File.separator;
			path += lp.getName()+TAIL;
			outFile = new File(path);
		}
		writeFile(outFile,lp.toString());		
	}
	public void save(ILightPattern<T> lp) {
		try {
			save2(lp);
		} catch(Exception e) {
			Log.getIstc().logln("Error: Can't store light pattern "+lp.getName()+".");
			Log.getIstc().logln(e.getMessage());
		}
	}
	public static void delete(ILightPattern lp) {
		try {
			File delFile = findFile(getUserPatternDir(),lp.getName());
			if(delFile!=null)
				delFile.delete();
		} catch(Exception e) {
			Log.getIstc().log("Error: Can't delete light pattern "+lp.getName()+".");
			Log.getIstc().logln(e.getMessage());
		}
	}
	public static void deleteCalculationRecords(String name) {
		try {
			File delFile = findFile(getUserRecordDir(),name);
			if(delFile!=null)
				delFile.delete();
		} catch(Exception e) {
			Log.getIstc().log("Error: Can't delete calculation record "+name+".");
			Log.getIstc().logln(e.getMessage());
		}
	}
	private static File findFile(String dr,String name) throws Exception {
		File dir = new File(dr);
		if(dir.exists()) {
			File[] files = dir.listFiles();
			if(files!=null&&files.length>0) {
				for(int i = 0;i<files.length;i++)
					if(files[i].getName().endsWith(TAIL)) {
						String fn = files[i].getName();
						String patName = fn.substring(0,fn.length()-TAIL.length());
						if(patName.compareTo(name)==0)
							return files[i];
					}
			}
		} else
			throw new Exception("Directory "+dir.getAbsolutePath()+" not found.");
		return null;
	}
	public String loadInfoText(String name) {
		String path = BASE_CONFIG_DIR+"/"+INFO_DIR+"/";
		return readResource(path+name+".html");
	}
	public LightChar[] loadLightChars() {
		String path = BASE_CONFIG_DIR+"/"+CHAR_DIR+"/";
		String fileContent = readResource(path+CHARS_FILE);
		StringTokenizer st = new StringTokenizer(fileContent,"\n");
		ArrayList<LightChar> lightChars = new ArrayList<LightChar>();
		LightChar lchar = null;
		int l = 0;
		try {
			while(st.hasMoreTokens()) {
				l++;
				String line = st.nextToken().trim();
				if(line.startsWith("#"))
					continue;
				int idx = line.indexOf("=");
				if(idx!=-1) {
					String key = line.substring(0,idx).trim();
					String val = line.substring(idx+1).trim();
					if(key.compareTo("char")==0) {
						String ch = val.substring(1,val.length()-1);
						lchar = new LightChar(ch);
						lightChars.add(lchar);
						if(ch.compareToIgnoreCase("bol")==0||ch.compareToIgnoreCase("eol")==0)
							lchar.setOrientation(LightChar.VERTICAL);
					} else if(key.equalsIgnoreCase("initial_gradients"))
						lchar.setInitialGradients(initialGradients(val));
					else if(key.equalsIgnoreCase("reflection_path"))
						lchar.setCellSequence(reflectionPath(val));
					else {
						throw new Exception("Unknown key: "+key);
					}
				}
			}
		} catch(Exception e) {
			String msg = "Error while reading file "+CHARS_FILE+TAIL+":\n";
			msg += "Can't read line "+l+".\n";
			Log.getIstc().log(msg);
			Log.getIstc().log(e);
			return null;
		}
		Collections.sort(lightChars,new Comparator<LightChar>() {

			@Override
			public int compare(LightChar s1,LightChar s2) {
				String c1 = s1.getChar();
				String c2 = s2.getChar();
				char ch1 = c1.charAt(0);
				char ch2 = c2.charAt(0);
				if(c1.length()==1&&c2.length()>1)
					return -1;
				if(c1.length()>1&&c2.length()==1)
					return 1;
				if(!Character.isLetter(ch1)&&Character.isLetter(ch2))
					return 1;
				if(Character.isLetter(ch1)&&!Character.isLetter(ch2))
					return -1;
				if(!Character.isDigit(ch1)&&Character.isDigit(ch2))
					return 1;
				if(Character.isDigit(ch1)&&!Character.isDigit(ch2))
					return -1;
				if(c1.length()==1&&c2.length()==1)
					return c1.compareTo(c2);
				return c1.compareTo(c2);
			}
		});
		return lightChars.toArray(new LightChar[lightChars.size()]);
	}
	public CalculationRecords<T> loadRecords(String name) {
		String dir = configDir+"/"+RECORD_DIR;
		String content;
		if(configDir.compareTo(BASE_CONFIG_DIR)==0)
			content = readResource(dir+"/"+name+TAIL);
		else
			content = loadText(name,dir,false);
		try {
			if(content!=null)
				return new CalculationRecords<T>(name,content,math);
		} catch (Exception e) {
			Log.getIstc().logln(e.getMessage());
		}
		return null;
	}
	public List<BaseLightPattern<T>> loadMuseum(List<BaseLightPattern<T>> li) {
		String dir = configDir+"/"+MUSEUM_DIR;
		String content;
		if(configDir.compareTo(BASE_CONFIG_DIR)==0)
			content = readResource(dir+"/"+MUSEUM_FILE);
		else
			content = loadText(new File(dir+"/"+MUSEUM_FILE),true);
		List<BaseLightPattern<T>> list = new ArrayList<BaseLightPattern<T>>();
		try {
			if(content!=null){
				StringTokenizer st = new StringTokenizer(content,"\n");
				while(st.hasMoreTokens()){
					String name = st.nextToken().trim();
					boolean found = false;
					for(BaseLightPattern<T> blp:li)
						if(blp.getName().compareTo(name) == 0){
							list.add(blp);
							found = true;
							break;
						}
					if(!found)
						Log.getIstc().logln(dir+"/"+MUSEUM_FILE+": Pattern "+name+" not found.");
				}
			}
				
		} catch (Exception e) {
			Log.getIstc().logln(e.getMessage());
		}
		return list;
	}
	public CalculationRecords<T> loadRecords(File f) {
		try {
			String txt = loadText(f,true);
			if(txt!=null)
				return new CalculationRecords<T>("File: "+f.getName(),txt,math);
		} catch (Exception e) {
			Log.getIstc().logln(e.getMessage());
		}
		Log.getIstc().logln("Error while reading file "+f.getAbsolutePath()+".");
		return null;
	}
	public List<ColorTheme> loadColorThemes() {
		String cfgList = readResource(BASE_CONFIG_DIR+"/"+CONFIG_LIST_FILE);
		if(cfgList!=null)
			return loadColorThemes(cfgList);
		File dir = new File(BASE_CONFIG_DIR+"/"+THEME_DIR);
		ArrayList<ColorTheme> retVal = new ArrayList<ColorTheme>();
		if(!dir.exists()) {
			Log.getIstc().log("Info: Can't load themes. ");
			Log.getIstc().logln("Directory "+dir.getAbsolutePath()+" not found.");
			return retVal;
		}
		File[] files = dir.listFiles();
		if(files!=null&&files.length>0) 
			for(int i = 0;i<files.length;i++){
				if(!files[i].getName().endsWith(TAIL)) 
					continue;
				String content;
				try {
					content = readFile(files[i]);
					if(content!=null&&content.length()>0) {
						int length = files[i].getName().length()-TAIL.length();
						String name = files[i].getName().substring(0,length);
						retVal.add(new ColorTheme(name,content));
					}
				} catch(Exception e) {
					String msg = "Warning: Can't read file ";
					Log.getIstc().log(msg+files[i].getAbsolutePath()+" .");
				}
			}
		return retVal;
	}
	public ColorTheme loadColorTheme(String theme) {
		if(theme!=null){
			List<ColorTheme> themes = loadColorThemes();
			for (ColorTheme t : themes) 
				if(t.getName().compareTo(theme)==0)
					return t;
		}
		return null;
	}
	public List<BaseLightPattern<T>> loadLightPatterns() {
		if(configDir.compareTo(BASE_CONFIG_DIR)==0){
			String path = BASE_CONFIG_DIR+"/"+CONFIG_LIST_FILE;
			String cfgList = readResource(path);
			if(cfgList!=null)
				return loadLightPatterns(cfgList);
		}
		File dir = new File(configDir+File.separator+PATTERN_DIR);
		List<BaseLightPattern<T>> retVal = new ArrayList<BaseLightPattern<T>>();
		if(!dir.exists()) {
			Log.getIstc().log("Info: Can't load light patterns. ");
			Log.getIstc().logln("Directory "+dir.getAbsolutePath()+" not found");
			return retVal;
		}
		File[] files = dir.listFiles();
		if(files!=null&&files.length>0) {
			LightChar[] lchs = loadLightChars();
			for(int i = 0;i<files.length;i++){
				if(!files[i].getName().endsWith(TAIL))
					continue;
				String content;
				try {
					content = readFile(files[i]);
					if(content!=null&&content.length()>0) {
						int length = files[i].getName().length()-TAIL.length();
						String name = files[i].getName().substring(0,length);
						BaseLightPattern<T> pat = buildPattern(name,content,lchs);
						if(pat!=null)
							retVal.add(pat);
					}
				} catch(IOException e) {
					Log.getIstc().log(e);
				}
			}
		}
		return retVal;
	}
	public BaseLightPattern<T> loadLightPattern(String pattern) {
		if(pattern!=null){
			List<BaseLightPattern<T>> pts = loadLightPatterns();
			for(BaseLightPattern<T> p:pts)
				if(p.getName().compareTo(pattern)==0)
					return p;
		}
		return null;
	}
	private static String loadText(String name,String dirName,boolean log) {
		File dir = new File(dirName);
		if(dir.exists()) {
			File file = new File(dirName+File.separator+name+TAIL);
			return loadText(file,log);
		} 
		Log.getIstc().logln("Error: Can't find directory: "+dir.getAbsolutePath());
		return null;
	}
	private static String loadText(File f,boolean log) {
		if(f.exists())
			try {
				return readFile(f);
			} catch(IOException e) {
				Log.getIstc().log(e);
			}
		else if(log)
			Log.getIstc().logln("Error: Can't find file: "+f.getAbsolutePath());
		return null;
	}
	private BaseLightPattern<T> buildPattern(String name,String fileContent,LightChar[] lc) {
		BaseLightPattern<T> retVal = null;
		StringTokenizer st = new StringTokenizer(fileContent,"\n");
		ArrayList<String> lines = new ArrayList<String>();
		while(st.hasMoreTokens()) {
			String s = st.nextToken();
			if(s.indexOf("=")!=-1)
				lines.add(s);
			else if(lines.size()>0)
				lines.set(lines.size()-1,lines.get(lines.size()-1)+"\n"+s);
		}
		for(String line:lines) {
			int idx = line.indexOf("=");
			String key = line.substring(0,idx).trim();
			String val = line.substring(idx+1).trim();
			if(retVal==null) {
				if(key.compareTo("text")==0) {
					try {
						retVal = new LightCharPattern<T>(null,null,false,lc,math);
					} catch(Exception e) {
						Log.getIstc().log(e);
					}
				} else
					retVal = new BaseLightPattern<T>(math);
			}
			if(key!=null&&key.length()>0)
				try {
					if(retVal instanceof LightCharPattern)
						put(key,val,(LightCharPattern<T>)retVal);
					else
						put(key,val,retVal);
				} catch(Exception e) {
					String msg = "Error while reading pattern file "+name+".txt"
					        +". Can't read line: "+line;
					Log.getIstc().logln(msg);
					Log.getIstc().log(e);
					return null;
				}
		}
		retVal.setName(name);
		retVal.calcStartIndex();
		if(retVal.getStartIndex()!=0)
			retVal.setPropagating(true);
		if(retVal.check())
			return retVal;
		return null;
	}
	private void put(String key,String val,LightCharPattern<T> pat) throws Exception {
		if(key.equalsIgnoreCase("text"))
			pat.setText(val);
		else if(key.equalsIgnoreCase("store_calc"))
			pat.setStoreCalculation(Boolean.parseBoolean(val));
		else
			throw new Exception("Unknown key: "+key);
	}
	private void put(String key,String val,BaseLightPattern<T> pat) throws Exception {
		if(key.equalsIgnoreCase("boundary_values")) {
			StringTokenizer st2 = new StringTokenizer(val,",");
			if(st2.countTokens()==2) {
				short maxPrec = Globals.PRECISIONS[Globals.PRECISIONS.length-1];
				T bndVal = math.util.parse(st2.nextToken().trim());
				pat.setStartBndVal(math.util.copy(bndVal,maxPrec));
				bndVal = math.util.parse(st2.nextToken().trim());
				pat.setEndBndVal(math.util.copy(bndVal,maxPrec));
			} else {
				String msg = "Two values are required. Found "+st2.countTokens();
				throw new Exception(msg);
			}
		} else if(key.equalsIgnoreCase("boundary_types")) {
			StringTokenizer st2 = new StringTokenizer(val,",");
			if(st2.countTokens()==2) {
				String s = st2.nextToken().trim();
				if(s.compareToIgnoreCase("p")==0)
					pat.setStartBndValType(BaseLightPattern.POINT);
				else if(s.compareToIgnoreCase("+g")==0)
					pat.setStartBndValType(BaseLightPattern.GRAD_RIGHT);
				else if(s.compareToIgnoreCase("-g")==0)
					pat.setStartBndValType(BaseLightPattern.GRAD_LEFT);
				else
					throw new Exception("Unknown bound type: "+s);
				s = st2.nextToken().trim();
				if(s.compareToIgnoreCase("p")==0)
					pat.setEndBndValType(BaseLightPattern.POINT);
				else if(s.compareToIgnoreCase("+g")==0)
					pat.setEndBndValType(BaseLightPattern.GRAD_RIGHT);
				else if(s.compareToIgnoreCase("-g")==0)
					pat.setEndBndValType(BaseLightPattern.GRAD_LEFT);
				else
					throw new Exception("Unknown bound type: "+s);
			} else
				throw new Exception("Expected two tokens");
		} else if(key.equalsIgnoreCase("initial_gradients"))
			pat.setInitialGradients(initialGradients(val));
		else if(key.equalsIgnoreCase("reflection_path"))
			pat.setCellSequence(reflectionPath(val));
		else if(key.equalsIgnoreCase("grid_size"))
			pat.setGridSize(Byte.parseByte(val));
		else if(key.equalsIgnoreCase("store_calc"))
			pat.setStoreCalculation(Boolean.parseBoolean(val));
		else if(key.equalsIgnoreCase("max_radius")) {
			if(val!=null&&val.length()>0) {
				int pos = val.indexOf("%");
				if(pos!=-1) {
					String percentValue = val.substring(0,pos).trim();
					short maxPrec = Globals.PRECISIONS[Globals.PRECISIONS.length-1];
					T r = math.util.copy(math.util.parse(percentValue),maxPrec);
					pat.setMaxShapeSize(r);
				} else
					throw new Exception("Missing percent symbol");
			} else
				pat.setMaxShapeSize(null);
		} else
			throw new Exception("Unknown key: "+key);
	}
	private static List<Float> initialGradients(String val) {
		StringTokenizer st2 = new StringTokenizer(val,",");
		List<Float> grads = new ArrayList<Float>();
		while(st2.hasMoreTokens())
			grads.add(Float.parseFloat(st2.nextToken()));
		return grads;
	}
	private static ArrayList<short[]> reflectionPath(String val) {
		StringTokenizer st2 = new StringTokenizer(val,",");
		ArrayList<short[]> cellList = new ArrayList<short[]>();
		while(st2.countTokens()>1) {
			String x = st2.nextToken();
			x = x.replace('(',' ').replace('{',' ').replace('[',' ').trim();
			String y = st2.nextToken();
			y = y.replace(')',' ').replace('}',' ').replace(']',' ').trim();
			short[] cellIdxs = new short[2];
			cellIdxs[0] = Byte.parseByte(x);
			cellIdxs[1] = Byte.parseByte(y);
			cellList.add(cellIdxs);
		}
		return cellList;
	}
	private List<ColorTheme> loadColorThemes(String resList){
		ArrayList<ColorTheme> retVal = new ArrayList<ColorTheme>();
		StringTokenizer st = new StringTokenizer(resList,"\n");
		while(st.hasMoreTokens()){
			String path = st.nextToken();
			if(path.indexOf("/"+THEME_DIR+"/")!=-1){
				try{
					String content = readResource(path);
					if(content==null||content.compareTo("")==0)
						throw new Exception();
					String name = path.substring(path.lastIndexOf("/")+1);
					name = name.substring(0,name.indexOf(TAIL));
					retVal.add(new ColorTheme(name,content));
				} catch(Exception e){
					Log.getIstc().log("Warning: Can't read file "+path+" .");
				}
			}
		}
		return retVal;
	}
	private List<BaseLightPattern<T>> loadLightPatterns(String resList){
		List<BaseLightPattern<T>> retVal = new ArrayList<BaseLightPattern<T>>();
		StringTokenizer st = new StringTokenizer(resList,"\n");
		LightChar[] lcs = loadLightChars();
		while(st.hasMoreTokens()){
			String path = st.nextToken();
			if(path.indexOf("/"+PATTERN_DIR+"/")!=-1){
				try{
					String content = readResource(path);
					if(content==null||content.compareTo("")==0)
						continue;
					String name = path.substring(path.lastIndexOf("/")+1);
					name = name.substring(0,name.indexOf(TAIL));
					retVal.add(buildPattern(name,content,lcs));
				} catch(Exception e){
					Log.getIstc().log("Warning: Pattern "+path+" not found.");
				}
			}
		}
		return retVal;
	}
	private String readResource(String name) {
		InputStream is = null;
		try{
			if((is=getClass().getResourceAsStream("/"+name))!=null){
			    BufferedInputStream bis = new BufferedInputStream(is);
			    ByteArrayOutputStream buf = new ByteArrayOutputStream();
			    int result;
			    while((result = bis.read())!= -1)
			    	buf.write((byte)result);
			    return buf.toString("UTF-8");
			}
		} catch(Exception e){
			Log.getIstc().logln("Error: Can't read resource "+name+" .");
			Log.getIstc().logln("Message: "+e.getMessage());
		} finally {
			if(is!=null)
				try { is.close(); } 
				catch (IOException e) {}
		}
		return null;
	}
	private static String readFile(File file) throws IOException {
		FileInputStream fis = new FileInputStream(file);
		try {
			byte[] bytes = new byte[(int)file.length()];
			fis.read(bytes);
			return new String(bytes,"UTF-8");
		} finally {
			fis.close();
		}
	}
	private static void writeFile(File file,String content) throws IOException {
		FileOutputStream fos = new FileOutputStream(file);
		try {
			fos.write(content.getBytes("UTF-8"));
		} finally {
			fos.close();
		}
	}
}
