/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package lattice;
import math.VecMath;

import common.IConverter;
import common.Log;

public abstract class Box<T extends Number> implements IBox<T> {

	public static final byte LOC_OUTSIDE = 0;
	public static final byte LOC_BOUNDARY = 1;
	public static final byte LOC_INSIDE = 2;
	protected T[] cornerPoint;
	protected T[][] baseVec;
	protected T[][] sideVec;
	protected T[] sideVecNorm;
	protected VecMath<T> math;
	public Box(T[] p,T length,VecMath<T> math) {
		this.cornerPoint = p;
		this.math = math;
		this.sideVec = math.util.array(p.length,p.length);
		this.baseVec = math.util.array(p.length,p.length);
		this.sideVecNorm = math.util.array(p.length);
		for(byte i = 0;i<sideVec.length;i++) {
			baseVec[i] = math.getBase(i,(byte)p.length).clone();
			sideVec[i] = baseVec[i].clone();
			math.scale(length,sideVec[i]);
			sideVecNorm[i] = length;
		}
	}
	public Box(T[] p,T[] lengths,VecMath<T> math) {
		this.cornerPoint = p;
		this.math = math;
		this.sideVec = math.util.array(lengths.length,p.length);
		this.baseVec = math.util.array(p.length,p.length);
		this.sideVecNorm = math.util.array((short)lengths.length);
		for(byte i = 0;i<baseVec.length;i++)
			baseVec[i] = math.getBase(i,(byte)p.length).clone();
		for(byte i = 0;i<sideVec.length;i++) {
			sideVec[i] = baseVec[i].clone();
			math.scale(lengths[i],sideVec[i]);
			sideVecNorm[i] = lengths[i];
		}
	}
	public Box(T[] p,T[][] baseVec,T[] lengths,VecMath<T> math) {
		this.cornerPoint = p;
		this.math = math;
		if(baseVec.length!=p.length) {
			String msg = "Base count does not match dimension";
			Log.getIstc().log(new Exception(msg));
			return;
		}
		this.baseVec = baseVec;
		this.sideVecNorm = lengths;
		short l1 = (short)lengths.length;
		short l2 = (short)p.length;
		this.sideVec = math.util.array(l1,l2);
		for(int i = 0;i<sideVec.length;i++)
			math.scale(baseVec[i],sideVecNorm[i],sideVec[i]);
	}
	/**
	 * Tests the location only in the sub space created by the
	 * side vectors.
	 */
	@Override
	public final byte location(T[] p) {
		T[] tmp = math.util.array(p.length);
		math.diff(p,cornerPoint,tmp);
		T[] param = math.util.array(p.length);
		math.decompose(tmp,baseVec,param);
		return locationLocal(param);
	}
	/**
	 * Assumes that the vector p is given in the base consisting of the 
	 * normalized side vectors. Tests the location only in the sub space 
	 * created by the side vectors. 
	 */
	@Override
	public final byte locationLocal(T[] p) {
		for(int i = 0;i<sideVecNorm.length;i++)
			if(math.cmp(math.ZERO,p[i])>0||math.cmp(sideVecNorm[i],p[i])<0)
				return LOC_OUTSIDE;
		for(int i = 0;i<sideVecNorm.length;i++)
			if(math.cmp(math.ZERO,p[i])==0||math.cmp(sideVecNorm[i],p[i])==0)
				return LOC_BOUNDARY;
		return LOC_INSIDE;
	}
	/**
	 * Tests the location only in the sub space created by the
	 * side vectors.
	 */
	@Override
	public boolean contains(T[] p) {
		byte loc = location(p);
		return loc==LOC_BOUNDARY||loc==LOC_INSIDE;
	}
	/**
	 * Assumes that the vector p is given in the base consisting of the 
	 * normalized side vectors. Tests the location only in the sub space 
	 * created by the side vectors. 
	 */
	@Override
	public boolean containsLocal(T[] p) {
		byte loc = locationLocal(p);
		return loc==LOC_BOUNDARY||loc==LOC_INSIDE;
	}
	@Override
	public final void constrainLocal(T[] p,T[] ret){
		for(int i = 0;i<p.length;i++){
			if(math.cmp(p[i],math.ZERO)<0)
				ret[i] = math.ZERO;
			else if(math.cmp(p[i],getLength((byte)i))>0)
				ret[i] = getLength((byte)i);
			else
				ret[i] = p[i];
		}
	}
	@Override
	public boolean intersect(Ray<T> ray,T[] isct) {
		T[] start = math.util.array(isct.length);
		T[] dir = math.util.array(isct.length);
		Ray<T> locRay = new Ray<T>(start,dir,math.util);
		world2LocalPoint(ray.getStart(),locRay.getStart());
		world2LocalVec(ray.getDir(),locRay.getDir());
		boolean isctFound = localIntersect(locRay,isct);
		if(isctFound)
			local2WorldPoint(isct.clone(),isct);
		return isctFound;
	}
	@Override
	public void setCorner(T[] corner) {
		math.util.copy(corner,this.cornerPoint);
	}
	@Override
	public void setCenter(T[] ctr) {
		math.scale(sideVecNorm,math.getDiagHalf((byte)ctr.length),cornerPoint);
		math.diff(ctr,cornerPoint,cornerPoint);
		setCorner(cornerPoint);
	}
	@Override
	public void setLength(T l) {
		for(int i = 0;i<sideVec.length;i++) {
			math.scale(baseVec[i],l,sideVec[i]);
			sideVecNorm[i] = l;
		}
	}
	@Override
	public void setLengths(T[] ls) {
		for(int i = 0;i<sideVec.length;i++) {
			math.scale(baseVec[i],ls[i],sideVec[i]);
			sideVecNorm[i] = ls[i];
		}
	}
	@Override
	public void setLength(T l,byte dir) {
		math.scale(baseVec[dir],l,sideVec[dir]);
		sideVecNorm[dir] = l;
	}
	@Override
	public final T getLength(byte dir) {
		return sideVecNorm[dir];
	}
	@Override
	public T[][] getSideVectors() {
		return sideVec;
	}
	@Override
	public T[][] getBaseVectors() {
		return baseVec;
	}
	@Override
	public T[] getLengths() {
		return sideVecNorm;
	}
	@Override
	public T[] getCorner() {
		return cornerPoint;
	}
	@Override
	public abstract IBox<T> copy();
	@Override
	public final void world2LocalPoint(T[] v,T[] ret) {
		T[] diff = math.util.array(v.length);
		math.diff(v,getCorner(),diff);
		math.decompose(diff,baseVec,ret);
	}
	@Override
	public final void local2WorldPoint(T[] v,T[] ret) {
		math.transform(v,baseVec,ret);
		math.add(getCorner(),ret,ret);
	}
	@Override
	public final void world2LocalVec(T[] v,T[] ret) {
		math.decompose(v,baseVec,ret);
	}
	@Override
	public final void local2WorldVec(T[] v,T[] ret) {
		math.transform(v,baseVec,ret);
	}
	public abstract <S extends Number>IBox<S> convert(IConverter<T,S> c,VecMath<S> smath);
}
