/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package lattice;
import math.VecMath;

import common.IConverter;

public class Box2D<T extends Number> extends Box<T> {

	private T[] locNorVec;
	public Box2D(T[] p1,T[][] baseVecs,T[] lengths,VecMath<T> math) {
		super(p1,baseVecs,lengths,math);
		init();
	}
	public Box2D(T[] p,T length,VecMath<T> math) {
		super(p,length,math);
		init();
	}
	public Box2D(T[] p,T[] lengths,VecMath<T> math) {
		super(p,lengths,math);
		init();
	}
	private void init() {
		if(cornerPoint.length>2) {
			this.locNorVec = math.ZERO3D.clone();
			this.locNorVec[2] = math.ONE;
		}
	}
	public boolean localIntersect(Ray<T> locRay,T[] isct) {
		// ray: r = start + t*dir;
		if(locNorVec!=null) {
			boolean ret = false;
			if(math.cmp(locRay.getDir()[2],math.ZERO)!=0)
				return localIntersect3D(locRay,isct);
			// locRay plane is parallel to this plane
			// check if locRay lies in this box2d plane
			if(math.cmp(locRay.getStart()[2],math.ZERO)==0)
				ret = localIntersect2D(locRay,isct);
			return ret;
		}
		return localIntersect2D(locRay,isct);
	}
	private boolean localIntersect3D(Ray<T> locRay,T[] isct) {
		T t = math.neg(math.div(locRay.getStart()[2],locRay.getDir()[2]));
		if(math.cmp(t,math.ZERO)<0)
			return false;
		math.scaleAdd(locRay.getStart(),t,locRay.getDir(),isct);
		return containsLocal(isct);
	}
	private boolean localIntersect2D(Ray<T> locRay,T[] isct) {
		T[] st = locRay.getStart();
		T[] dir = locRay.getDir();
		byte loc = locationLocal(st);
		if(loc==LOC_BOUNDARY) {
			math.util.copy(locRay.getStart(),isct);
			return true;
		}
		for(int i = 0;i<2;i++) {
			if(math.cmp(dir[1-i],math.ZERO)!=0) {
				math.util.copy(math.ZERO3D,isct);
				boolean firstSide = math.cmp(dir[1-i],math.ZERO)>0;
				if(loc==LOC_OUTSIDE) {
					if(firstSide) {
						if(math.cmp(st[1-i],math.ZERO)>0)
							continue;
					} else if(math.cmp(st[1-i],sideVecNorm[1-i])<0)
						continue;
				} else
					firstSide = !firstSide;
				short sideIdx = (short)(i+(firstSide?0:2));
				switch(sideIdx) {
					case 0:
						isct[0] = math.sub(st[0],math.mul(st[1],math.div(dir[0],dir[1])));
						isct[1] = math.ZERO;
						break;
					case 1:
						isct[0] = math.ZERO;
						isct[1] = math.sub(st[1],math.mul(st[0],math.div(dir[1],dir[0])));
						break;
					case 2:
						T tmp = math.sub(sideVecNorm[1],st[1]);
						isct[0] = math.add(st[0],math.mul(tmp,math.div(dir[0],dir[1])));
						isct[1] = sideVecNorm[1];
						break;
					case 3:
						tmp = math.sub(sideVecNorm[0],st[0]);
						isct[0] = sideVecNorm[0];
						isct[1] = math.add(st[1],math.mul(tmp,math.div(dir[1],dir[0])));
				}
				if(containsLocal(isct))
					return true;
			}
		}
		return false;
	}
	public T getWidth() {
		return getLength((byte)0);
	}
	public T getHeight() {
		return getLength((byte)1);
	}
	@Override
	public void rotate(T[] p,T[] axis,T phi) {
		math.rotate2D(cornerPoint,p,phi);
		for(int i = 0;i<sideVec.length;i++) {
			math.rotate2D(baseVec[i],p,phi);
			math.scale(baseVec[i],sideVecNorm[i],sideVec[i]);
		}
	}
	@Override
	public void rotate(T[] p,T[] axis,T sinphi,T cosphi) {
		math.rotate2D(cornerPoint,p,sinphi,cosphi);
		for(int i = 0;i<sideVec.length;i++) {
			math.rotate2D(baseVec[i],p,sinphi,cosphi);
			math.scale(baseVec[i],sideVecNorm[i],sideVec[i]);
		}
	}
	@Override
	public Box2D<T> copy() {
		T[][] bv = baseVec.clone();
		for(int i = 0;i<bv.length;i++)
			bv[i] = baseVec[i].clone();
		T[] svn = sideVecNorm.clone();
		T[] cp = cornerPoint.clone();
		return new Box2D<T>(cp,bv,svn,math);
	}
	@Override
	public <S extends Number>Box2D<S> convert(IConverter<T,S> c,VecMath<S> smath) {
		S[][] bv = smath.util.array((short)baseVec.length,(short)baseVec[0].length);
		for(int i = 0;i<bv.length;i++)
			for(int j = 0;j<bv.length;j++)
				bv[i][j] = c.convert(baseVec[i][j]);
		S[] svn = c.convert(sideVecNorm);
		S[] cp = c.convert(cornerPoint);
		return new Box2D<S>(cp,bv,svn,smath);
	}
}
