/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package lattice;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import math.Util;
import math.VecMath;

import common.IConverter;
import common.Log;

@SuppressWarnings("serial")
public final class RayPolygon<T extends Number> extends ArrayList<T[]> {

	private final Util<T> util;
	private final double sqrt_8 = Math.sqrt(8);
	private String description;
	private Ray<T> inRay;
	private Ray<T> outRay;
	private IShape<T> shape;
	private final ArrayList<Float> digitLossHigh;
	private final ArrayList<Float> digitLossLow;
	private int accuracy;
	private double qualityBound;
	private boolean digitLossQuality;
	private boolean calcDigitLoss;
	private boolean obfuscated;
	// Variables used by the loop count algorithm
	private final float[] tmpPoint;
	private float[] coords;
	private float[] inStart;
	private float[] inDir;
	private float[] outStart;
	private float[] outDir;
	private int size;
	public RayPolygon(Util<T> util,int accuracy) {
		this.util = util;
		this.calcDigitLoss = true;
		this.digitLossHigh = new ArrayList<Float>();
		this.digitLossLow = new ArrayList<Float>();
		this.coords = new float[100];
		this.tmpPoint = new float[2];
		this.size = 0;
		setAccuracy(accuracy);
	}
	public void setAccuracy(int a) {
		this.accuracy = a;
		this.qualityBound = Math.pow(10,-Math.min(accuracy-1,15));
	}
	public int getAccuracy() {
		return accuracy;
	}
	@Override
	public boolean add(T[] pt) {
		boolean b = super.add(pt);
		if(outRay==null)
			incDigitLoss();
		else if(calcDigitLoss)
			rebuild();
		return b;
	}
	public void setInRay(Ray<T> in) {
		this.inRay = in;
		if(digitLossLow.size()>0)
			rebuild();
	}
	public void setOutRay(Ray<T> out) {
		if(outRay!=null)
			if(calcDigitLoss) {
				digitLossLow.remove(digitLossLow.size()-1);
				digitLossHigh.remove(digitLossHigh.size()-1);
			}
		this.outRay = out;
		if(outRay!=null)
			incDigitLoss();
	}
	@Override
	public synchronized void clear() {
		inRay = null;
		outRay = null;
		digitLossHigh.clear();
		digitLossLow.clear();
		digitLossQuality = true;
		obfuscated = false;
		super.clear();
	}
	@Override
	public void add(int pos,T[] pt) {
		super.add(pos,pt);
		if(calcDigitLoss)
			rebuild();
	}
	@Override
	public boolean addAll(Collection<? extends T[]> pts) {
		boolean b = true;
		if(outRay!=null) {
			b = super.addAll(pts);
			if(calcDigitLoss)
				rebuild();
		} else {
			Iterator<? extends T[]> iter = pts.iterator();
			while(iter.hasNext())
				b &= add(iter.next());
		}
		return b;
	}
	@Override
	public synchronized T[] remove(int pos) {
		T[] retVal = super.remove(pos);
		if(calcDigitLoss)
			rebuild();
		return retVal;
	}
	public synchronized void removeLast() {
		if(size()>0) {
			super.remove(size()-1);
			if(calcDigitLoss)
				if(outRay==null) {
					digitLossHigh.remove(digitLossHigh.size()-1);
					digitLossLow.remove(digitLossLow.size()-1);
				} else
					rebuild();
		}
	}
	@Override
	public synchronized boolean remove(Object arg0) {
		Exception e = new Exception("Error: RayPolygon.remove(Object) not supported");
		e.fillInStackTrace();
		Log.getIstc().logln(e.getMessage());
		return false;
	}
	@Override
	public synchronized void removeRange(int pos1,int pos2) {
		boolean end = pos2>=size();
		super.removeRange(pos1,pos2);
		if(calcDigitLoss)
			if(end&&outRay==null)
				for(int i = pos2-1;i>=pos1;i--) {
					digitLossLow.remove(i);
					digitLossHigh.remove(i);
				}
			else
				rebuild();
	}
	private void rebuild() {
		ArrayList<T[]> pts = new ArrayList<T[]>(this);
		Ray<T> ray1 = inRay;
		Ray<T> ray2 = outRay;
		clear();
		setInRay(ray1);
		Iterator<T[]> iter = pts.iterator();
		while(iter.hasNext())
			add(iter.next());
		setOutRay(ray2);
	}
	public Ray<T> getInRay() {
		return inRay;
	}
	public Ray<T> getOutRay() {
		return outRay;
	}
	public void setShape(IShape<T> s) {
		this.shape = s;
	}
	public float getDigitLossHigh() {
		return digitLossHigh.size()>0?getDigitLossHigh(digitLossHigh.size()-1):0;
	}
	public float getDigitLossLow() {
		return digitLossLow.size()>0?getDigitLossLow(digitLossLow.size()-1):0;
	}
	public float getDigitLossHigh(int pos) {
		return digitLossHigh.get(pos);
	}
	public float getDigitLossLow(int pos) {
		return digitLossLow.get(pos);
	}
	public String getDescription() {
		return description;
	}
	public void setDescription(String description) {
		this.description = description;
	}
	public void setErrorCalculation(boolean b) {
		this.calcDigitLoss = b;
	}
	public boolean getDigitLossQuality() {
		return digitLossQuality;
	}
	public void setObfuscated(boolean b) {
		obfuscated = b;
	}
	public boolean isObfuscated() {
		return obfuscated;
	}
	public synchronized float calcLength(IBox<Float> fbox,IConverter<T,Float> cv,
			VecMath<Float> fmath) {
		float retVal = 0;
		float length = fbox.getLength((byte)0);
		if(!obfuscated&&getInRay()!=null) {
			float[] crds = util.cast2floatArray(this);
			for(int i = 1;i<crds.length/2;i++) {
				float dx = crds[2*i]-crds[2*(i-1)];
				float dy = crds[2*i+1]-crds[2*(i-1)+1];
				retVal += Math.sqrt(dx*dx+dy*dy);
			}
			Ray<Float> r = getInRay().convert(cv,fmath.util);
			Float[] p1 = new Float[2];
			if(fbox.intersect(r,p1)) {
				Float[] p2 = new Float[2];
				if(size()==0) {
					if(fbox.location(r.getStart())==Box.LOC_INSIDE)
						fmath.diff(p1,r.getStart(),p2);
					else {
						Float[] st = r.getStart();
						fmath.scaleAdd(st,2*length,r.getDir(),st);
						fmath.neg(r.getDir());
						fbox.intersect(r,p2);
						fmath.diff(p1,p2,p2);
					}
				} else {
					if(getOutRay()!=null) {
						r = getOutRay().convert(cv,fmath.util);
						fbox.intersect(r,p2);
						int l = size()-1;
						fmath.diff(util.cast2FloatArray(get(l)),p2,p2);
						retVal += fmath.norm(p2);
					}
					fmath.diff(util.cast2FloatArray(get(0)),p1,p2);
				}
				retVal += fmath.norm(p2);
			}
		}
		return retVal/length;
	}
	public int countRays() {
		int cnt = inRay!=null?1:0;
		cnt += outRay!=null?1:0;
		return cnt+Math.max(2*size()/2-1,0);
	}
	public synchronized int countLoops() {
		int loops = 0;
		if(inRay!=null){
			float[] st1 = util.cast2floatArray(inRay.getStart());
			float[] dir1 = util.cast2floatArray(inRay.getDir());
			float[] st2 = null;
			float[] dir2 = null;
			if(outRay!=null) {
				st2 = util.cast2floatArray(outRay.getStart());
				dir2 = util.cast2floatArray(outRay.getDir());
			}
			loops = countLoops(st1,dir1,util.cast2floatArray(this),st2,dir2);
		}
		return loops;
	}
	private void incDigitLoss() {
		if(calcDigitLoss)
			if(shape instanceof Circle<?>)
				incDigitLossUseCircle(shape.getSize().doubleValue());
			else {
				String n = shape.getClass().getName();
				Log.getIstc().logln("No method to calculate errors for "+n);
			}
	}
	private void incDigitLossUseCircle(double r) {
		double[] dirVec1 = new double[2];
		double[] dirVec2 = new double[2];
		int i = outRay==null?size()-2:size()-1;
		if(i>0||i==0&&inRay!=null) {
			double l1;
			if(i>0) {
				dirVec1[0] = get(i-1)[0].doubleValue()-get(i)[0].doubleValue();
				dirVec1[1] = get(i-1)[1].doubleValue()-get(i)[1].doubleValue();
				l1 = Math.sqrt(dirVec1[0]*dirVec1[0]+dirVec1[1]*dirVec1[1]);
				double inv_l1 = 1/l1;
				dirVec1[0] *= inv_l1;
				dirVec1[1] *= inv_l1;
			} else {
				l1 = 1;
				dirVec1[0] = -getInRay().getDir()[0].doubleValue();
				dirVec1[1] = -getInRay().getDir()[1].doubleValue();
			}
			double l2;
			if(i==size()-2) {
				dirVec2[0] = get(i+1)[0].doubleValue()-get(i)[0].doubleValue();
				dirVec2[1] = get(i+1)[1].doubleValue()-get(i)[1].doubleValue();
				l2 = Math.sqrt(dirVec2[0]*dirVec2[0]+dirVec2[1]*dirVec2[1]);
				double inv_l2 = 1/l2;
				dirVec2[0] *= inv_l2;
				dirVec2[1] *= inv_l2;
			} else {
				l2 = 1000000;
				dirVec2[0] = outRay.getDir()[0].doubleValue();
				dirVec2[1] = outRay.getDir()[1].doubleValue();
			}
			double dot = dirVec1[0]*dirVec2[0]+dirVec1[1]*dirVec2[1];
			if(dot+1<qualityBound) {
				dot = qualityBound-1;
				digitLossQuality = false;
			}
			double tmp = sqrt_8*l1/(r*Math.sqrt(1+dot));
			int lastPos = digitLossHigh.size()-1;
			float dlHigh = digitLossHigh.get(lastPos);
			float dlLow = digitLossLow.get(lastPos);
			digitLossHigh.add(dlHigh+(float)Math.log10((l1+l2)/l2+tmp));
			digitLossLow.add(dlLow+(float)Math.log10(1+tmp));
		} else {
			digitLossHigh.add(0f);
			digitLossLow.add(0f);
		}
	}
	private int countLoops(float[] st1,float[] dir1,float[] crds,float[] st2,
	        float[] dir2) {
		int cnt = 0;
		if(crds.length>2) {
			init(st1,dir1,crds,st2,dir2);
			int i = 1;
			while(i<size/2-1) {
				float[] pt = null;
				for(int k = i-2;k>=0;k--) {
					pt = intersect(2*k,2*i);
					if(pt!=null) {
						cnt++;
						insertAndRemove(2*(k+1),2*i,pt);
						i = k;
						break;
					}
				}
				if(pt==null) {
					pt = intersect(inStart,inDir,2*i);
					if(pt!=null) {
						cnt++;
						insertAndRemove(0,2*i,pt);
						i = 0;
					}
				}
				i++;
			}
			if(outStart!=null) {
				for(int k = size/2-3;k>=0;k--) {
					float[] pt = intersect(2*k,outStart,outDir);
					if(pt!=null)
						++cnt;
				}
				float[] pt = intersect(inStart,inDir,outStart,outDir);
				if(pt!=null)
					++cnt;
			}
		}
		return cnt;
	}
	private void init(float[] st1,float[] dir1,float[] crds,float[] st2,float[] dir2) {
		coords = crds;
		size = crds.length;
		if(size>0) {
			if(inStart==null)
				inStart = new float[2];
			inStart[0] = coords[0];
			inStart[1] = coords[1];
			if(inDir==null)
				inDir = new float[2];
			inDir[0] = -dir1[0];
			inDir[1] = -dir1[1];
		} else {
			inStart = st1;
			inDir = dir1;
		}
		outStart = st2;
		outDir = dir2;
	}
	private float[] intersect(int idx1,int idx2) {
		float dir1X = coords[idx1+2]-coords[idx1];
		float dir1Y = coords[idx1+3]-coords[idx1+1];
		if(dir1X!=0||dir1Y!=0) {
			float norDir1X = -dir1Y;
			float norDir1Y = dir1X;
			float dir2X = coords[idx2+2]-coords[idx2];
			float dir2Y = coords[idx2+3]-coords[idx2+1];
			if(dir2X!=0||dir2Y!=0) {
				float dot = norDir1X*dir2X+norDir1Y*dir2Y;
				if(dot!=0) {
					float t2 = (coords[idx1]-coords[idx2])*norDir1X;
					t2 += (coords[idx1+1]-coords[idx2+1])*norDir1Y;
					t2 /= dot;
					// <1 ensures correctness if two reflection points are equal
					if(t2>=0&&t2<1) {
						tmpPoint[0] = coords[idx2]+t2*dir2X;
						tmpPoint[1] = coords[idx2+1]+t2*dir2Y;
						float t1 = (tmpPoint[0]-coords[idx1])*dir1X;
						t1 += (tmpPoint[1]-coords[idx1+1])*dir1Y;
						if(t1>0&&t1<=sqNorm(dir1X,dir1Y))
							return tmpPoint;
					}
				}
				// Parallel case needs no extra handling because of law of reflection
			}
		}
		return null;
	}
	private float[] intersect(float[] st1,float[] dir1,float[] st2,float[] dir2) {
		float dot = -dir1[1]*dir2[0];
		dot += dir1[0]*dir2[1];
		if(dot!=0) {
			float t2 = (st1[0]-st2[0])*-dir1[1];
			t2 += (st1[1]-st2[1])*dir1[0];
			t2 /= dot;
			if(t2>=0) {
				tmpPoint[0] = st2[0]+t2*dir2[0];
				tmpPoint[1] = st2[1]+t2*dir2[1];
				float t1 = (tmpPoint[0]-st1[0])*dir1[0];
				t1 += (tmpPoint[1]-st1[1])*dir1[1];
				if(t1>=0)
					return tmpPoint;
			}
		}
		return null;
	}
	private float[] intersect(float[] st,float[] dir,int idx) {
		float dir2X = coords[idx+2]-coords[idx];
		float dir2Y = coords[idx+3]-coords[idx+1];
		if(dir2X!=0||dir2Y!=0) {
			float dot = -dir[1]*dir2X+dir[0]*dir2Y;
			if(dot!=0) {
				float t2 = (st[0]-coords[idx])*-dir[1];
				t2 += (st[1]-coords[idx+1])*dir[0];
				t2 /= dot;
				if(t2>=0&&t2<1) {
					tmpPoint[0] = coords[idx]+t2*dir2X;
					tmpPoint[1] = coords[idx+1]+t2*dir2Y;
					float t1 = (tmpPoint[0]-st[0])*dir[0];
					t1 += (tmpPoint[1]-st[1])*dir[1];
					if(t1>=0)
						return tmpPoint;
				}
			}
		}
		return null;
	}
	private float[] intersect(int idx,float[] st,float[] dir) {
		float dir1X = coords[idx+2]-coords[idx];
		float dir1Y = coords[idx+3]-coords[idx+1];
		if(dir1X!=0||dir1Y!=0) {
			float dot = -dir1Y*dir[0]+dir1X*dir[1];
			if(dot!=0) {
				float t2 = (coords[idx]-st[0])*-dir1Y;
				t2 += (coords[idx+1]-st[1])*dir1X;
				t2 /= dot;
				if(t2>=0) {
					tmpPoint[0] = st[0]+t2*dir[0];
					tmpPoint[1] = st[1]+t2*dir[1];
					float t1 = (tmpPoint[0]-coords[idx])*dir1X;
					t1 += (tmpPoint[1]-coords[idx+1])*dir1Y;
					if(t1>0&&t1<=sqNorm(dir1X,dir1Y))
						return tmpPoint;
				}
			}
		}
		return null;
	}
	private static float sqNorm(float c1,float c2) {
		return c1*c1+c2*c2;
	}
	private void insertAndRemove(int i1,int i2,float[] pt) {
		if(i1==0) {
			inStart[0] = pt[0];
			inStart[1] = pt[1];
		}
		coords[i1] = pt[0];
		coords[i1+1] = pt[1];
		System.arraycopy(coords,i2+2,coords,i1+2,size-i2-2);
		size -= i2-i1;
	}
}
