/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package lattice;
import java.util.ArrayList;

import math.VecMath;

import common.Globals;
import common.ICallback;
import common.Log;

public class ReflectionEngine<T extends Number> {
	
	public final static byte OUTPUT_LOCAL = 0;
	public final static byte OUTPUT_WORLD = 1;
	private final static int CALC_INIT_PREC = 15;
	private final static int CALC_INC_PREC = 2;
	private final VecMath<T> mt;
	private final int iniAc;
	private final int incAc;
	private final ArrayList<T[]> tmpCoords;
	private final RayPolygon<T> locCoords;
	private final T[] outputPt;
	private final T[] reflPt;
	private boolean calc = true;
	private byte outputType;
	public ReflectionEngine(VecMath<T> mt){
		this(mt,CALC_INIT_PREC,CALC_INC_PREC);
	}
	public ReflectionEngine(VecMath<T> mt,int ac_start,int ac_inc) {
		this.mt = mt;
		this.iniAc = ac_start;
		this.incAc = ac_inc;
		this.tmpCoords = new ArrayList<T[]>();
		this.locCoords = new RayPolygon<T>(mt.util,-1);
		this.outputPt = mt.util.array(2);
		this.reflPt = mt.util.array(2);
		this.outputType = OUTPUT_WORLD;
	}
	/**
	 * Calculates the first reflection points very fast, but gets slower and
	 * slower the more points have to be calculated.
	 */
	public synchronized boolean calcDynamic(Ray<T> ray,ShapeContainer<T> ctr,
			RayPolygon<T> poly,ICallback cb) {
		tmpCoords.clear();
		locCoords.clear();
		locCoords.setAccuracy(poly.getAccuracy());
		poly.clear();
		poly.setErrorCalculation(true);
		locCoords.setShape(ctr.getShape());
		poly.setShape(ctr.getShape());
		int limitAc = mt.PRECISION-ray.getScale()-2;
		int maxAc = Math.min(ray.getAccuracy(),limitAc);
		if(maxAc<limitAc) {
			if(Globals.LOG_VERBOSE) {
				Log.getIstc().log("ReflectionEngine: Warning: Ray accuracy "+maxAc);
				Log.getIstc().logln(" is lower than max accuracy "+limitAc);
			}
			if(maxAc<iniAc*incAc) {
				Log.getIstc().log("ReflectionEngine: Error: ");
				Log.getIstc().logln("Minimum ray accuracy must be "+iniAc*incAc);
			}
		}
		boolean cvrgc = true;
		ShapeContainer<T>.LocalReflIter iter = ctr.new LocalReflIter();
		int[] levelAcs = getLevelAccuracies(maxAc);
		byte level = 0;
		do {
			Ray<T> calcRay = new Ray<T>(ray,levelAcs[level]+ray.getScale()+2,mt.util);
			iter.dispose();
			iter.init(calcRay,false);
			boolean outsideShape = poly.size()>0||!iter.isStartingInShape();
			if(level==0) {
				Ray<T> locInRay = new Ray<T>(iter.getInitRay(),-1,mt.util);
				if(outputType==OUTPUT_WORLD)
					poly.setInRay(calcRay);
				else
					poly.setInRay(locInRay);
				locCoords.setInRay(locInRay);
			}
			if(!outsideShape) {
				cvrgc = false;
				poly.setObfuscated(true);
				break;
			} else if(!iter.hasNext())
				break;
			cvrgc = calcReflPts(ctr,poly,cb,levelAcs[level],level==levelAcs.length-1,iter);
		} while((calc||poly.size()==0)&&!cvrgc&&++level<levelAcs.length);
		if(calc&&cvrgc) {
			Ray<T> reflRay = iter.getReflRay();
			Ray<T> r = new Ray<T>(reflRay,poly.getAccuracy(),mt.util);
			if(outputType==OUTPUT_WORLD) {
				ctr.local2WorldPoint(reflRay.getStart(),r.getStart());
				ctr.local2WorldVec(reflRay.getDir(),r.getDir());
			}
			poly.setOutRay(r);
		}
		if(Globals.LOG_VERBOSE) {
			String state = poly.getDigitLossQuality()?"ok":"only min";
			String msg = "ReflectionEngine: Finished calculation"+"\n  max digit loss: "
			        +poly.getDigitLossHigh()+"\n  min digit loss: "+poly.getDigitLossLow()
			        +"\n  digit loss calculation: "+state;
			Log.getIstc().logln(msg);
		}
		iter.dispose();
		return cvrgc;
	}
	private boolean calcReflPts(ShapeContainer<T> ctr,RayPolygon<T> poly,ICallback cb,int ac,
	        boolean last,ShapeContainer<T>.LocalReflIter it) {
		boolean cvrgc;
		cvrgc = tmpCoords.size()!=0;
		if(cvrgc)
			locCoords.removeRange(poly.size(),locCoords.size());
		int reqAc = poly.getAccuracy();
		int stAc = it.getInitRay().getAccuracy();
		for(int k = 0;(calc||ac<=iniAc*incAc)&&it.hasNext()&&stAc-locCoords.getDigitLossLow()>reqAc;k++) {
			it.next(reflPt);
			if(k>=poly.size()) {
				mt.util.round(reflPt,poly.getAccuracy());
				cvrgc &= tmpCoords.size()!=0;
				if(cvrgc) {
					T[] pt = tmpCoords.remove(0);
					cvrgc = mt.cmp(pt[0],reflPt[0])==0&&mt.cmp(pt[1],reflPt[1])==0;
					if(!cvrgc) {
						tmpCoords.clear();
						if(last)
							break;
						callback(cb);
					}
				}
				if(cvrgc) {
					if(outputType==OUTPUT_WORLD)
						ctr.local2WorldPoint(reflPt,outputPt);
					else
						mt.util.copy(reflPt,outputPt);
					poly.add(outputPt.clone());
				} else
					tmpCoords.add(reflPt.clone());
				locCoords.add(reflPt.clone());
				it.setAccuracy((int)Math.ceil(stAc-locCoords.getDigitLossLow()));
			} else
				it.setAccuracy((int)Math.ceil(stAc-locCoords.getDigitLossLow(k)));
		}
		if(Globals.LOG_VERBOSE) {
			String msg = "ReflectionEngine:"+"\n  accuracy: "+stAc+"\n  verified points: "
			        +poly.size()+"\n  cached points: "+tmpCoords.size()+"\n  max digit loss: "
			        +locCoords.getDigitLossHigh()+"\n  min digit loss: "
			        +locCoords.getDigitLossLow();
			Log.getIstc().logln(msg);
		}
		return cvrgc;
	}
	/**
	 * Calculates the first reflection points slowly, but gets faster and
	 * faster the more points have to be calculated.
	 */
	public synchronized boolean calcStatic(Ray<T> ray,ShapeContainer<T> ctr,
			RayPolygon<T> poly,ICallback cb) {
		poly.clear();
		poly.setErrorCalculation(true);
		poly.setShape(ctr.getShape());
		ShapeContainer<T>.LocalReflIter iter = ctr.new LocalReflIter();
		iter.init(ray,false);
		Ray<T> locInRay = new Ray<T>(iter.getInitRay(),-1,mt.util);
		if(outputType==OUTPUT_WORLD)
			poly.setInRay(ray);
		else
			poly.setInRay(locInRay);
		boolean cvrgc = !iter.isStartingInShape();
		if(cvrgc) {
			int stAc = ray.getAccuracy();
			int reqAc = poly.getAccuracy();
			// TODO Many order error increment near end may be undetected.
			// Switch to upper calc method if digit loss quality is low
			while(!(cvrgc = !iter.hasNext())&&(calc||poly.size()==0)
			        &&stAc-poly.getDigitLossHigh()>reqAc+2) {
				iter.next(reflPt);
				mt.util.round(reflPt,reqAc+2);
				if(outputType==OUTPUT_WORLD)
					ctr.local2WorldPoint(reflPt,outputPt);
				else
					mt.util.copy(reflPt,outputPt);
				poly.add(outputPt.clone());
				iter.setAccuracy((int)Math.ceil(stAc-poly.getDigitLossLow()));
				if(poly.size()%25==0)
					callback(cb);
			}
			if(cvrgc) {
				Ray<T> reflRay = iter.getReflRay();
				Ray<T> r = new Ray<T>(reflRay,poly.getAccuracy(),mt.util);
				if(outputType==OUTPUT_WORLD) {
					ctr.local2WorldPoint(reflRay.getStart(),r.getStart());
					ctr.local2WorldVec(reflRay.getDir(),r.getDir());
				}
				poly.setOutRay(r);
			} else {
				poly.removeLast();
				poly.removeLast();
			}
		} else
			poly.setObfuscated(true);
		if(Globals.LOG_VERBOSE) {
			String state = poly.getDigitLossQuality()?"ok":"only min";
			String msg = "ReflectionEngine: Finished calculation"+"\n  max digit loss: "
			        +poly.getDigitLossHigh()+"\n  min digit loss: "+poly.getDigitLossLow()
			        +"\n  digit loss calculation: "+state;
			Log.getIstc().logln(msg);
		}
		iter.dispose();
		// Log.getIstc().logln("Static calculation.");
		return cvrgc;
	}
	private int[] getLevelAccuracies(int maxAccu) {
		float tmp = (float)(Math.log10(maxAccu*1.0/iniAc)/Math.log10(incAc));
		int levels = Math.round(tmp)+2;
		int[] retVal = new int[levels];
		for(int j = 0,ac = iniAc;j<retVal.length;j++,ac *= incAc) {
			if(j==retVal.length-2)
				retVal[j] = ac-2*j++;
			retVal[j] = ac;
		}
		return retVal;
	}
	private static void callback(ICallback cb) {
		if(cb!=null)
			cb.call();
	}
	public void setStopped(boolean b) {
		this.calc = !b;
	}
	public void setOutputType(byte ot) {
		this.outputType = ot;
	}
}
