/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package lattice;
import java.util.Iterator;
import java.util.LinkedList;

import math.VecMath;

import common.IConverter;

public final class ShapeContainer<T extends Number> extends Lattice<T> {

	// Local variables often used
	private final LinkedList<ShapeContainer<T>.LocalCellIterator> cellIterList;
	private final LinkedList<IShape<T>> shapeList;
	private final LinkedList<Ray<T>> rayList;
	// The shape to be put into the grid cells
	private IShape<T> defaultShape;
	public ShapeContainer(T[] p,short sidePoints,T gap,VecMath<T> math) {
		super(p,sidePoints,gap,math);
		cellIterList = new LinkedList<ShapeContainer<T>.LocalCellIterator>();
		shapeList = new LinkedList<IShape<T>>();
		rayList = new LinkedList<Ray<T>>();
	}
	public ShapeContainer(ShapeContainer<T> ctr,VecMath<T> math) {
		this(ctr.box.copy(),ctr.getPoints(),ctr.getGap(),math,ctr.getShape().copy());
		//this.defaultShape = ctr.getShape().copy();
	}
	private ShapeContainer(IBox<T> box,short pts,T gap,VecMath<T> math,IShape<T> s) {
		super(box,pts,gap,math);
		defaultShape = s;
//		cellIterList = null;
//		shapeList = null;
//		rayList = null;
		cellIterList = new LinkedList<ShapeContainer<T>.LocalCellIterator>();
		shapeList = new LinkedList<IShape<T>>();
		rayList = new LinkedList<Ray<T>>();
	}
	public void setShape(IShape<T> s) {
		this.defaultShape = s;
		clearLocalVars();
	}
	public IShape<T> getShape() {
		return this.defaultShape;
	}
	public IShape<T> getShape(T[] pt) {
		IShape<T> retVal = this.defaultShape.copy();
		getCellCenter(pt,retVal.getCenter());
		return retVal.contains(pt)?retVal:null;
	}
	public IShape<T> getShapeLocal(T[] locPt) {
		IShape<T> retVal = this.defaultShape.copy();
		getCellCenterLocal(locPt,retVal.getCenter());
		return retVal.contains(locPt)?retVal:null;
	}
	@Override
	public void rotate(T[] center,T[] axis,T phi) {
		T[] sinCos = math.sincos(phi);
		rotate(center,axis,sinCos[0],sinCos[1]);
	}
	@Override
	public void rotate(T[] center,T[] axis,T sinphi,T cosphi) {
		super.rotate(center,axis,sinphi,cosphi);
		defaultShape.rotate(defaultShape.getCenter(),null,sinphi,cosphi);
		clearLocalVars();
	}
	@Override
	public void setGap(T gp) {
		defaultShape.scale(math.div(gp,getGap()));
		super.setGap(gp);
		clearLocalVars();
	}
	@Override
	public void setPoints(short pts) {
		defaultShape.scale(math.util.cast(getPoints()-1));
		defaultShape.scale(math.inv(math.util.cast(pts-1)));
		super.setPoints(pts);
		T size = math.util.copy(math.util.cast(pts-1),math.PRECISION);
		super.setGap(math.div(getLengths()[0],size));
		clearLocalVars();
	}
	private void clearLocalVars() {
		if(shapeList!=null)
			shapeList.clear();
		if(rayList!=null)
			rayList.clear();
		if(cellIterList!=null)
			cellIterList.clear();
	}
	@Override
	public <S extends Number>ShapeContainer<S> convert(IConverter<T,S> cv,VecMath<S> m) {
		IBox<S> sbox = super.convert(cv,m);
		S sGap = cv.convert(getGap());
		IShape<S> sShape = null;
		if(defaultShape!=null)
			sShape = defaultShape.convert(cv,m);
		return new ShapeContainer<S>(sbox,getPoints(),sGap,m,sShape);
	}
	public class ShapeIteratorFloat implements Iterator<IShape<Float>> {

		private final IShape<Float> shape;
		private final Float[][] bv;
		private final IndexIterator idxIter;
		private final Float[] cellCtrCor;
		private final VecMath<Float> fmath;
		public ShapeIteratorFloat(IConverter<T,Float> cv,VecMath<Float> fmt) {
			fmath = fmt;
			float gap = getGap().floatValue();
			T[][] bvT = getBaseVectors().clone();
			bv = new Float[bvT.length][bvT[0].length];
			for(int i = 0;i<bv.length;i++) {
				for(int j = 0;j<bv[0].length;j++)
					bv[i][j] = bvT[i][j].floatValue()*gap;
			}
			this.idxIter = new IndexIterator();
			shape = getShape().convert(cv,fmath);
			short[] dIdxs = new short[bv.length];
			for(byte i = 0;i<dIdxs.length;i++)
				dIdxs[i] = 1;
			cellCtrCor = fmath.util.array(2);
			fmath.linBaseComb(dIdxs,bv,cellCtrCor);
			fmath.scale(fmath.ONE_OVER_TWO,cellCtrCor);
			Float[] cor = cv.convert(getCorner());
			fmath.add(cor,cellCtrCor,cellCtrCor);
		}
		public boolean hasNext() {
			return idxIter.hasNext();
		}
		public IShape<Float> next() {
			if(hasNext()) {
				short[] idxs = idxIter.next();
				Float[] cellCtr = cellCtrCor.clone();
				fmath.linBaseCombAdd(idxs,bv,cellCtr);
				shape.setCenter(cellCtr);
				return shape;
			}
			return null;
		}
		public void remove() {}
	}
	public class ShapeIterator implements Iterator<IShape<T>> {

		private final IShape<T> shape;
		private final T[][] bv;
		private final IndexIterator idxIter;
		private final T [] cellCtrCor;
		public ShapeIterator(int prec) {
			T gap = math.util.copy(getGap(),prec);
			this.bv = getBaseVectors().clone();
			for(int i = 0;i<bv.length;i++) {
				bv[i] = bv[i].clone();
				for(int j = 0;j<bv[0].length;j++)
					bv[i][j] = math.mul(bv[i][j],gap);
			}
			this.idxIter = new IndexIterator();
			shape = getShape().copy();

			short[] dIdxs = new short[bv.length];
			for(byte i = 0;i<dIdxs.length;i++)
				dIdxs[i] = 1;
			cellCtrCor = math.util.array(2);
			math.linBaseComb(dIdxs,bv,cellCtrCor);
			math.scale(math.ONE_OVER_TWO,cellCtrCor);
			math.add(getCorner(),cellCtrCor,cellCtrCor);
		}
		public boolean hasNext() {
			return idxIter.hasNext();
		}
		public IShape<T> next() {
			if(hasNext()) {
				short[] idxs = idxIter.next();
				T[] cellCtr = cellCtrCor.clone();
				math.linBaseCombAdd(idxs,bv,cellCtr);
				shape.setCenter(cellCtr);
				return shape;
			}
			return null;
		}
		public void remove() {}
	}
	public class LocalReflIter implements Iterator<T[]> {

		private ShapeContainer<T>.LocalCellIterator cellItr;
		private IShape<T> shape;
		private Ray<T> reflRay;
		private Ray<T> initRay;
		private boolean foundNext;
		private boolean empty = true;
		private int accuracy;
		private short[] startCellIdx;
		public void init(Ray<T> ray,boolean all) {
			this.shape = shapeList.pollFirst();
			if(shape==null)
				shape = defaultShape.copy();
			T diff = math.sub(getGap(),math.mul(shape.getSize(),math.TWO));
			int scale = util.getScale(diff);
			if(all)
				startCellIdx = null;
			this.cellItr = cellIterList.pollFirst();
			if(cellItr==null)
				cellItr = ShapeContainer.this.new LocalCellIterator();
			cellItr.init(ray,scale,startCellIdx);
			startCellIdx = cellItr.getStartCellIndex();
			startCellIdx = startCellIdx!=null?startCellIdx.clone():null;
			this.reflRay = rayList.pollFirst();
			if(reflRay==null)
				reflRay = new Ray<T>(math.util);
			reflRay.copy(cellItr.getLocalRay());
			if(initRay==null)
				initRay = new Ray<T>(math.util);
			initRay.copy(cellItr.getLocalRay());
			this.foundNext = false;
			this.empty = false;
			accuracy = -1;
		}
		@Override
		public boolean hasNext() {
			if(empty)
				return false;
			if(!foundNext) {
				while(!foundNext&&cellItr.hasNext()) {
					cellItr.next(shape.getCenter());
					foundNext = shape.reflect(reflRay,reflRay);
				}
				if(foundNext)
					cellItr.continueWithRay(reflRay);
				else
					empty = true;
			}
			return foundNext;
		}
		public boolean isStartingInShape() {
			boolean retVal = false;
			if(cellItr!=null&&cellItr.hasNext()) {
				Ray<T> locRay = cellItr.getLocalRay();
				cellItr.next(shape.getCenter());
				if(shape.contains(locRay.getStart()))
					retVal = true;
				cellItr.replay();
			}
			return retVal;
		}
		@Override
		public T[] next() {
			T[] pt = util.array(2);
			boolean res = next(pt);
			if(!res)
				pt = null;
			return pt;
		}
		public boolean next(T[] pt) {
			if(!empty)
				if(foundNext||hasNext()) {
					foundNext = false;
					util.copy(reflRay.getStart(),pt);
					return true;
				}
			return false;
		}
		@Override
		public void remove() {}
		public Ray<T> getInitRay() {
			return initRay;
		}
		public Ray<T> getReflRay() {
			return reflRay;
		}
		public void setAccuracy(int digits) {
			accuracy = digits;
			int rayAccu = reflRay.getAccuracy();
			if(rayAccu<accuracy)
				reflRay.incPrecision(accuracy-rayAccu+5);
		}
		public void dispose() {
			if(shape!=null) {
				shapeList.add(shape);
				shape = null;
			}
			if(cellItr!=null) {
				cellIterList.add(cellItr);
				cellItr = null;
			}
			if(reflRay!=null) {
				rayList.add(reflRay);
				reflRay = null;
			}
		}
	}
}
