/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import lattice.Box2D;
import lattice.IBox;
import lattice.IShape;
import lattice.Ray;
import lattice.RayPolygon;
import lattice.ReflectionEngine;
import lattice.ShapeContainer;
import math.Util;
import math.VecMath;
import math.VecMathDouble;
import math.VecMathFloat;
import mvc.scan.ScanListener;
import pattern.BaseLightPattern;
import pattern.CalculationRecords;
import pattern.ILightPattern;
import pattern.LightChar;

import common.DoubleConverter;
import common.FileIO;
import common.FloatConverter;
import common.Globals;
import common.ICallback;
import common.IConverter;
import common.Log;

public final class Model<T extends Number> implements ICallback {

	// Geometry
	public static final int WORLD_SIDE_LENGTH = 3;
	public final static int GRID_SIDE_LENGTH = 2;
	public final static int GRID_CORNER_X = -1;
	public final static int GRID_CORNER_Y = -1;
	// Events
	public static final byte GRID_ROTATE = 0;
	public static final byte GRID_RESIZE = 5;
	public static final byte SHAPE_RESIZE = 8;
	public static final byte RAY_NEW = 10;
	public static final byte RAY_UPDATE = 20;
	public static final byte RAY_FINISHED = 30;
	public static final byte MODEL_NEW = 40;
	public static final byte MODEL_WORLD_RESIZE = 50;
	public static final byte LIGHTPATTERNS_UPDATE = 60;
	// Calculation mode
	public final static byte DYNAMIC_CALC = 0;
	public final static byte STATIC_CALC = 1;
	// Direction of the light source ray
	public final T[] LIGHT_SOURCE_DIR;
	// The box containing the whole screen
	private final Box2D<T> worldBox;
	// The box containing the unrotated circle grid
	private final IBox<T> ctrInitBox;
	// Arbitrary precision math
	private final VecMath<T> tmath;
	// Float precision math
	private final VecMath<Float> fmath;
	// Double precision math
	private final VecMath<Double> dmath;
	private final IConverter<T,Float> tfltCvtr;
	// Contains the circles arranged to a grid
	private final ShapeContainer<T> tShapeCtr;
	private final ReflectionEngine<T> calculator;
	// The ray starting from the light source
	private Ray<T> ray;
	// The rotation angle of the shape container
	// TODO: Put in ShapeContainer
	private T angle;
	// Contains in and out going rays and the reflection points
	private final RayPolygon<T> rayPolygon;
	// Contains reflection points that have been calculated
	// after the last listener call
	private final ArrayList<T> rayDataUpdate;
	// The listeners to be called if new reflection points are
	// available or if the grid or model changes
	private final LinkedList<IModelListener> modelListeners;
	// The listeners to be called if new scan data is available
	private final LinkedList<ScanListener> scanListeners;
	// Number of reflection points for which a listener update has been made
	private int lastSize = 0;
	// Contains all data needed for a scan run
	private ScanSession<T> scanSession;
	// Indicates if listeners should be called
	private boolean update;
	// Dynamic or static
	private byte calcMode;
	public Model(T[] cor,short pts,T gap,T[] wh,VecMath<T> mt) {
		this.LIGHT_SOURCE_DIR = mt.util.array(2);
		this.LIGHT_SOURCE_DIR[0] = mt.util.cast(-1);
		this.LIGHT_SOURCE_DIR[1] = mt.ZERO;
		this.tmath = mt;
		this.fmath = new VecMathFloat();
		this.dmath = new VecMathDouble();
		this.tfltCvtr = new FloatConverter<T>();
		this.tShapeCtr = new ShapeContainer<T>(cor,pts,gap,tmath);
		this.calculator = new ReflectionEngine<T>(tmath);
		this.ctrInitBox = tShapeCtr.copy();
		T[] boxcor = tmath.util.array(2);
		boxcor[0] = tmath.mul(tmath.neg(wh[0]),tmath.ONE_OVER_TWO);
		boxcor[1] = tmath.mul(tmath.neg(wh[1]),tmath.ONE_OVER_TWO);
		this.worldBox = new Box2D<T>(boxcor,wh,tmath);
		this.angle = mt.ZERO;
		this.modelListeners = new LinkedList<IModelListener>();
		this.scanListeners = new LinkedList<ScanListener>();
		this.rayPolygon = new RayPolygon<T>(mt.util,Globals.CALC_ACCURACY);
		this.rayDataUpdate = new ArrayList<T>();
		this.update = true;
	}
	@Override
	public void call() {
		checkForRayDataUpdate(false);
	}
	public void setRay(Ray<T> r,String descr,byte cm) {
		ray = r;
        T[] st = ray.getStart().clone();
        tmath.diff(st,worldBox.getCorner(),st);
        if(!worldBox.containsLocal(st)){
        	worldBox.constrainLocal(st,st);
        	tmath.add(st,worldBox.getCorner(),st);
        	ray.setStart(st);
        }
		calcMode = cm;
		clearRayData();
		rayPolygon.setDescription(descr);
		fireEvent(RAY_NEW);
	}
	public void setRay(Ray<T> r,byte cm) {
		setRay(r,null,cm);
	}
	public Ray<T> getRay() {
		return ray;
	}
	public String getDescription() {
		return rayPolygon.getDescription();
	}
	public void calculateReflections() {
		if(ray!=null){
			ICallback cb = update?this:null;
			switch(calcMode) {
				case STATIC_CALC:
					calculator.calcStatic(ray,tShapeCtr,rayPolygon,cb);
					break;
				case DYNAMIC_CALC:
					calculator.calcDynamic(ray,tShapeCtr,rayPolygon,cb);
			}
			checkForRayDataUpdate(true);
		}
	}
	public void setCalcMode(byte cm) {
		this.calcMode = cm;
	}
	void setCalculatorActive(boolean b) {
		calculator.setStopped(!b);
	}
	public IBox<T> getGridInitBound() {
		return this.ctrInitBox;
	}
	public T getGridGap() {
		return tShapeCtr.getGap();
	}
	public short getGridSize() {
		return (short)(tShapeCtr.getPoints()-1);
	}
	public void setGridSize(short i) {
		if(tShapeCtr.getPoints()-1!=i) {
			clearRayData();
			tShapeCtr.setPoints(++i);
			fireEvent(GRID_RESIZE);
		}
	}
	public T calcGap(short gridSize) {
		T s = tmath.util.copy(tmath.util.cast(gridSize),tmath.PRECISION);
		return tmath.div(tShapeCtr.getLengths()[0],s);
	}
	public int getDecimalPlaces() {
		if(!rayPolygon.isObfuscated()&&rayPolygon.getOutRay()==null)
			return Util.MAX_INT;
		return (int)Math.ceil(rayPolygon.getDigitLossHigh())+3;
	}
	public void addModelListener(IModelListener ml) {
		synchronized(modelListeners){
			modelListeners.add(ml);
		}
	}
	public void addScanListener(ScanListener sl) {
		synchronized(scanListeners) {
			scanListeners.add(sl);
		}
	}
	public void removeScanListener(ScanListener l) {
		synchronized(scanListeners) {
			scanListeners.remove(l);
		}
	}
	public void removeModelListener(IModelListener ml) {
		synchronized(modelListeners){
			modelListeners.remove(ml);
		}
	}
	public boolean isInGrid(T[] point) {
		return tShapeCtr.contains(point);
	}
	public void rotateGrid(T phi) {
		clearRayData();
		T[] sinCos = tmath.sincos(phi);
		rotateGrid(sinCos,tShapeCtr);
		angle = tmath.add(angle,phi);
		fireEvent(GRID_ROTATE);
	}
	public void rotateGrid(T phi,T[] sinCosPhi) {
		clearRayData();
		rotateGrid(sinCosPhi,tShapeCtr);
		angle = tmath.add(angle,phi);
		fireEvent(GRID_ROTATE);
	}
	public void rotateGrid(T[] sinCosPhi,ShapeContainer<T> ctr) {
		ctr.rotate(tmath.ZERO2D,null,sinCosPhi[0],sinCosPhi[1]);
	}
	public T getGridRotation() {
		return angle;
	}
	public void setGridRotation(T phi) {
		clearRayData();
		setGridRotation(phi,tShapeCtr);
		angle = phi;
		fireEvent(GRID_ROTATE);
	}
	public void setGridRotation(T phi,T[] sincosphi) {
		clearRayData();
		setGridRotation(sincosphi,tShapeCtr);
		angle = phi;
		fireEvent(GRID_ROTATE);
	}
	private void setGridRotation(T phi,ShapeContainer<T> ctr) {
		ctr.setBound(ctrInitBox.copy());
		ctr.rotate(tmath.ZERO2D,null,phi);
	}
	private void setGridRotation(T[] sincosphi,ShapeContainer<T> ctr) {
		ctr.setBound(ctrInitBox.copy());
		ctr.rotate(tmath.ZERO2D,null,sincosphi[0],sincosphi[1]);
	}
	public void setShape(IShape<T> s) {
		clearRayData();
		tShapeCtr.setShape(s);
		fireEvent(SHAPE_RESIZE);
	}
	public IShape<T> getShape() {
		return tShapeCtr.getShape();
	}
	public IShape<T> getShape(T[] pt) {
		return tShapeCtr.getShape(pt);
	}
	public void setShapePercentSize(byte percent) {
		setShapePercentSize(tmath.util.cast(percent));
	}
	public void setShapePercentSize(T percent) {
		clearRayData();
		IShape<T> s = getShape();
		s.setSize(transformShapeSize(percent));
		setShape(s);
	}
	public T transformShapeSize(byte percent) {
		return transformShapeSize(tmath.util.cast(percent));
	}
	public T transformShapeSize(T percent) {
		return tmath.div(tmath.mul(getGridGap(),percent),tmath.util.cast(200));
	}
	public T getShapePercentSize() {
		T s = getShape().getSize();
		s = tmath.div(tmath.mul(s,tmath.util.cast(200)),getGridGap());
		int pos = -tmath.util.getErrorScale(s);
		if(pos<Util.MAX_INT)
			return tmath.util.round(s,pos-1);
		return s;
	}
	public byte getShapeSizeIdx() {
		for(byte i = 0;i<Globals.SHAPE_SIZES.length;i++) {
			T size = transformShapeSize(tmath.util.cast(Globals.SHAPE_SIZES[i]));
			if(tmath.cmp(size,getShape().getSize())==0)
				return i;
		}
		return -1;
	}
	public T calcShapeSize(short gridSize) {
		IShape<T> shape = getShape().copy();
		shape.scale(tmath.util.cast(getGridSize()));
		shape.scale(tmath.inv(tmath.util.cast(gridSize)));
		return shape.getSize();
	}
	public ShapeContainer<Double> getShapeContainerDouble(){
		return tShapeCtr.convert(new DoubleConverter<T>(),dmath);
	}
	public Iterator<IShape<Float>> getShapeIteratorFloat() {
		return tShapeCtr.new ShapeIteratorFloat(tfltCvtr,fmath);
	}
	public Box2D<T> getWorldBox() {
		return worldBox;
	}
	public boolean updateWorldBox(Box2D<T> world) {
		boolean retVal = false;
		if(this.worldBox!=world) {
			this.worldBox.setCorner(world.getCorner());
			this.worldBox.setLengths(world.getLengths());
			if(ray != null){
		        T[] st = ray.getStart();
		        if(!worldBox.contains(st)){
		        	setRay(ray,rayPolygon.getDescription(),STATIC_CALC);
		        	retVal = true;
		        }
			}
		}
		fireEvent(MODEL_WORLD_RESIZE);
		return retVal;
	}
	public VecMath<T> getMath() {
		return tmath;
	}
	public VecMath<Float> getFloatMath() {
		return fmath;
	}
	public VecMath<Double> getDoubleMath() {
		return dmath;
	}
	public ScanSession<T> getScanSession() {
		return scanSession;
	}
	public synchronized void getRayDataComplete(LinkedList<T> list) {
		if(rayPolygon.getInRay()!=null) {
			list.add(rayPolygon.getInRay().getStart()[0]);
			list.add(rayPolygon.getInRay().getStart()[1]);
			for(int i = 0;(1+Math.max(2*i/2-1,0)<lastSize)&&
					(i<rayPolygon.size());i++) {
				list.add(rayPolygon.get(i)[0]);
				list.add(rayPolygon.get(i)[1]);
			}
			T[] point = tmath.util.array(2);
			if(rayPolygon.getOutRay()!=null) {
				rayPolygon.getOutRay().getPoint(tmath.TEN,point,tmath);
				list.add(point[0]);
				list.add(point[1]);
			}
		}
	}
	public synchronized void getRayDataUpdate(LinkedList<T> upd) {
		upd.addAll(rayDataUpdate);
	}
	private synchronized void checkForRayDataUpdate(boolean calcFinished) {
		int rayCount = rayPolygon.countRays();
		if(lastSize<rayCount) {
			rayDataUpdate.clear();
			if(lastSize==0)
				if(rayPolygon.getInRay()!=null) {
					rayDataUpdate.add(rayPolygon.getInRay().getStart()[0]);
					rayDataUpdate.add(rayPolygon.getInRay().getStart()[1]);
				}
			for(int i = Math.max((lastSize-2),0);i<rayPolygon.size();i++) {
				rayDataUpdate.add(rayPolygon.get(i)[0]);
				rayDataUpdate.add(rayPolygon.get(i)[1]);
			}
			if(rayPolygon.getOutRay()!=null) {
				T[] sPoint = tmath.util.array(2);
				T[] lgs = worldBox.getLengths();
				T param = tmath.cmp(lgs[0],lgs[1])>=0?lgs[0]:lgs[1];
				param = tmath.mul(param,tmath.TWO);
				rayPolygon.getOutRay().getPoint(param,sPoint,tmath);
				rayDataUpdate.add(sPoint[0]);
				rayDataUpdate.add(sPoint[1]);
			}
			lastSize = rayCount;
			fireEvent(RAY_UPDATE);
		}
		if(calcFinished)
			fireEvent(RAY_FINISHED);
	}
	private void fireEvent(byte event) {
		synchronized(modelListeners){
			for(IModelListener ml:modelListeners)
				ml.modelChanged(event);
		}
	}
	public void initScanSession(PixelMap<T> pmap,ScanSession<T> se) {
		ShapeContainer<T> ctr = new ShapeContainer<T>(tShapeCtr,tmath);
		ctr.setBound(this.ctrInitBox.copy());
		T[] sinCosPhiStart = tmath.sincos(pmap.xPix2World((short)0));
		Log.getIstc().logln("Rotating scan grid to phi start");
		ctr.rotate(tmath.ZERO2D,tmath.ZERO2D,sinCosPhiStart[0],sinCosPhiStart[1]);
		se.tCtr = ctr;
		scanSession = se;
	}
	private void clearRayData() {
		rayPolygon.clear();
		rayDataUpdate.clear();
		lastSize = 0;
	}
	public void updateWhileCalc(boolean upd) {
		this.update = upd;
	}
	public List<IModelListener> getModelListeners() {
		return modelListeners;
	}
	public List<ScanListener> getScanListeners() {
		return scanListeners;
	}
	public int getReflectionCount() {
		int retVal = Math.max(rayPolygon.size(),0);
		return rayPolygon.getOutRay()==null?-retVal:retVal;
	}
	public float getLength() {
		IBox<Float> fbox = tShapeCtr.copy().convert(tfltCvtr,fmath);
		float retVal = rayPolygon.calcLength(fbox,tfltCvtr,fmath);
		return rayPolygon.getOutRay()==null&&retVal!=0?-retVal:retVal;
	}
	public int getLoops() {
		int retVal = rayPolygon.countLoops();
		return rayPolygon.getOutRay()==null&&retVal!=0?-retVal:retVal;
	}
	public List<BaseLightPattern<T>> getLightPatterns() {
		return new FileIO<T>(tmath).loadLightPatterns();
	}	
	public List<BaseLightPattern<T>> getMuseumPatterns() {
		return new FileIO<T>(tmath).loadMuseum(getLightPatterns());
	}
	public void addLightPattern(ILightPattern<T> lp) {
		new FileIO<T>(tmath).save(lp);
		fireEvent(LIGHTPATTERNS_UPDATE);
	}
	public void addRecords(CalculationRecords<T> rec,File f){
		new FileIO<T>(tmath).save(rec,f);
	}
	public void addRecords(CalculationRecords<T> rec) {
		new FileIO<T>(tmath).save(rec);
	}
	public CalculationRecords<T> getRecords(String patternName) {
		return new FileIO<T>(tmath).loadRecords(patternName);
	}
	public CalculationRecords<T> getRecords(File f) {
		return new FileIO<T>(tmath).loadRecords(f);
	}
	public void deleteRecords(String patternName) {
		FileIO.deleteCalculationRecords(patternName);
	}
	public CalculationRecords<T> createRecord(String name,int p) {
		short s = getGridSize();
		CalculationRecords<T> retVal = new CalculationRecords<T>(name,s,tmath);
		T spz = getShapePercentSize();
		T[] start = getRay().getStart().clone();
		T phi = getGridRotation();
		if(p!=-1){
			start[1] = tmath.util.copy(start[1],p);
			phi = tmath.util.copy(phi,p);
		}
		retVal.add(spz,tmath.util.round(start[0],4),start[1],phi);
		return retVal;
	}
	public void setRecord(CalculationRecords<T>.Record rec,
			                         short size) {
		if(size!=-1)
			setGridSize(size);
		T phi = tmath.util.copy(rec.getAngle(),tmath.PRECISION);
		setGridRotation(phi);
		if(rec.getShapeSize()!=null){
			T s = tmath.util.copy(rec.getShapeSize(),tmath.PRECISION);
			setShapePercentSize(s);
		}
		T[] start = tmath.util.array(2);
		start[0] = tmath.util.copy(rec.getDistance(),tmath.PRECISION);
		start[1] = tmath.util.copy(rec.getHeight(),tmath.PRECISION);
		Ray<T> r = new Ray<T>(start,LIGHT_SOURCE_DIR,tmath.util);
		String name = rec.getEnclosingType().getPatternName();
		setRay(r,name,Model.STATIC_CALC);
	}
	public void removeLightPattern(ILightPattern<T> lp) {
		FileIO.delete(lp);
		FileIO.deleteCalculationRecords(lp.getName());
		fireEvent(LIGHTPATTERNS_UPDATE);
	}
	public LightChar[] getLightChars() {
		return new FileIO<T>(tmath).loadLightChars();
	}
}
