/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc;
import java.awt.Dimension;
import java.awt.GraphicsEnvironment;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.List;

import javax.swing.JFrame;
import javax.swing.WindowConstants;

import lattice.IBox;
import mvc.scan.ScanMenuBar;
import mvc.scan.ScanVisualizer;
import mvc.var.VariableMenuBar;
import mvc.var.VariableVisualizer;
import pattern.BaseLightPattern;

import common.ColorTheme;
import common.ICallback;

@SuppressWarnings("serial")
public final class View<T extends Number> extends JFrame {

	private final static String TITLE_BASE = "FroZenLight";
	private final Controller<T> ctrl;
	private final Visualizer<T> reflVis;
	private ScanVisualizer<T> scanVis;
	private VariableVisualizer<T> varVis;
	private MainMenuBar<T> menuBar;
	private LightPatternSelector<T> lpSel;
	public View(final Controller<T> ctrl,ColorTheme ct,ICallback cb,byte win) {
		super(TITLE_BASE);
		Rectangle rect = GraphicsEnvironment.getLocalGraphicsEnvironment()
		.getMaximumWindowBounds();
		boolean scrs = win==Controller.WIN_SCREENSAVER;
		boolean muf = win==Controller.WIN_MUSEUM_FULL;
		if(win==Controller.WIN_FULLSCREEN||scrs||muf){
			this.setUndecorated(true);
			this.setResizable(false);
		} else 
			rect = new Rectangle(Math.min(650,rect.width),Math.min(500,rect.height));
		this.ctrl = ctrl;
		this.reflVis = new Visualizer<T>(ctrl,ct,cb,scrs);
		this.menuBar = new MainMenuBar<T>(this,ct);
		boolean mu = muf || win==Controller.WIN_MUSEUM;
		if(mu){
			List<BaseLightPattern<T>> lps = ctrl.getModel().getMuseumPatterns();
			reflVis.setTouchOverlay(new TouchButtonsOverlay<T>(lps,menuBar));
		}
		initGUI(ct,!(scrs||mu),rect);
		ctrl.getModel().addModelListener(new IModelListener() {

			@Override
			public void modelChanged(byte event) {
				switch(event) {
					case Model.RAY_NEW:
						if(getTitle().indexOf("-")!=-1)
							setTitle(TITLE_BASE);
						break;
					case Model.RAY_FINISHED:
						String descr = ctrl.getModel().getDescription();
						if(descr!=null&&descr.compareTo("")!=0)
							setTitle(TITLE_BASE+" - "+descr);
				}
			}
		});
	}
	private void initGUI(ColorTheme ct,boolean mb,Rectangle rect) {
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		getContentPane().setLayout(gridbag);
		c.fill = GridBagConstraints.BOTH;
		
//		c.weightx = 0.1;
//		c.gridx = 0;
//		c.gridy = 0;
//		c.gridwidth = 1;
//		c.gridheight = 2;		
//		TouchMenu tm = new TouchMenu();
//		gridbag.setConstraints(tm,c);
//		getContentPane().add(tm);
//		
//		c.gridx = 2;
//		c.gridy = 0;
//		c.gridwidth = 1;
//		c.gridheight = 2;		
//		tm = new TouchMenu();
//		gridbag.setConstraints(tm,c);
//		getContentPane().add(tm);	
		
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 1;
		c.gridheight = 1;
		if(mb){
			Dimension d = menuBar.getPreferredSize();
			menuBar.setMinimumSize(new Dimension(10,d.height));
			gridbag.setConstraints(menuBar,c);
			getContentPane().add(menuBar);
			c.gridy = 1;
		}
		c.weightx = 1;
		c.weighty = 1;
		gridbag.setConstraints(reflVis,c);
		getContentPane().add(reflVis);
		setDefaultCloseOperation(EXIT_ON_CLOSE);
		pack();
		setSize(rect.width,rect.height);
		setVisible(true);
		reflVis.paintImmediately(reflVis.getVisibleRect());
		if(menuBar!=null)
			menuBar.paintImmediately(reflVis.getVisibleRect());
	}
	public Visualizer<T> getReflectionVisualizer() {
		return reflVis;
	}
	public JFrame createScanView() {
		if(scanVis==null) {
			scanVis = new ScanVisualizer<T>(ctrl);
			JFrame scanViewFrame = new JFrame("Scanner");
			GridBagLayout gridbag = new GridBagLayout();
			GridBagConstraints c = new GridBagConstraints();
			scanViewFrame.getContentPane().setLayout(gridbag);
			c.fill = GridBagConstraints.BOTH;
			c.gridx = 0;
			c.gridy = 0;
			c.weightx = 1;
			c.weighty = 0;
			c.gridwidth = 1;
			c.gridheight = 1;
			final ScanMenuBar<T> scanMenuBar = new ScanMenuBar<T>(scanVis);
            Dimension d = scanMenuBar.getPreferredSize();
			scanMenuBar.setMinimumSize(new Dimension(40,d.height));
			gridbag.setConstraints(scanMenuBar,c);
			scanViewFrame.getContentPane().add(scanMenuBar);
			c.fill = GridBagConstraints.BOTH;
			c.gridx = 0;
			c.gridy = 1;
			c.weightx = 1;
			c.weighty = 1;
			c.gridwidth = 1;
			c.gridheight = 1;
			gridbag.setConstraints(scanVis,c);
			scanViewFrame.getContentPane().add(scanVis);
			scanViewFrame.addWindowListener(new WindowAdapter() {

				@Override
				public void windowClosed(WindowEvent e) {
					closeAction();
				}
				@Override
				public void windowClosing(WindowEvent e) {
					closeAction();
				}
				private void closeAction() {
					ctrl.stopScan("Stopped on close");
					if(scanVis!=null){
						scanVis.onDelete();
						scanMenuBar.onDelete();
						scanVis = null;
					}
					if(varVis!=null)
						varVis.setScannerTabEnabled(false);
				}
			});
			scanViewFrame.pack();
			scanViewFrame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
			scanViewFrame.setVisible(true);
			scanMenuBar.postInit();
			return scanViewFrame;
		}
		return null;
	}
	public ScanVisualizer<T> getScanVisualizer() {
		return scanVis;
	}
	public VariableVisualizer<T> getVariableVisualizer() {
		return varVis;
	}
	public Controller<T> getController() {
		return ctrl;
	}
	public JFrame createVariableView() {
		if(varVis==null) {
			varVis = new VariableVisualizer<T>(ctrl);
			if(scanVis!=null) {
				varVis.setScannerTabEnabled(true);
				IBox<T> box = scanVis.getPixelMap().getWorldBox();
				varVis.updateScannerFrameFields(box);
				varVis.updateScannerObsFields(scanVis.getObservables());
			}
			JFrame varViewFrame = new JFrame("Variables");
			GridBagLayout gridbag = new GridBagLayout();
			GridBagConstraints c = new GridBagConstraints();
			varViewFrame.getContentPane().setLayout(gridbag);
			c.fill = GridBagConstraints.BOTH;
			c.gridx = 0;
			c.gridy = 0;
			c.weightx = 1;
			c.weighty = 0;
			c.gridwidth = 1;
			c.gridheight = 1;
			final VariableMenuBar<T> vmb = new VariableMenuBar<T>(varVis);
			Dimension d = vmb.getPreferredSize();
			vmb.setMinimumSize(new Dimension(10,d.height));
			gridbag.setConstraints(vmb,c);
			varViewFrame.getContentPane().add(vmb);
			c.fill = GridBagConstraints.BOTH;
			c.insets = new Insets(10,10,10,10);
			c.gridx = 0;
			c.gridy = 1;
			c.weightx = 1;
			c.weighty = 1;
			c.gridwidth = 1;
			c.gridheight = 1;
			gridbag.setConstraints(varVis,c);
			varViewFrame.getContentPane().add(varVis);
			varViewFrame.addWindowListener(new WindowAdapter() {

				@Override
				public void windowClosed(WindowEvent e) {
					if(varVis!=null){
						varVis.onDelete();
						vmb.onDelete();
						varVis = null;
					}
				}
				@Override
				public void windowClosing(WindowEvent e) {
					if(varVis!=null){
						varVis.onDelete();
						vmb.onDelete();
						varVis = null;
					}
				}
			});
			varViewFrame.pack();
			varViewFrame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
			varViewFrame.setVisible(true);
			return varViewFrame;
		}
		return null;
	}
	MainMenuBar<T> getMainMenuBar() {
		return menuBar;
	}
	public LightPatternSelector<T> getLightPatternSelector() {
		if(lpSel==null)
			lpSel = new LightPatternSelector<T>(this);
		return lpSel;
	}
	public void setLightSource(short lsType){
		reflVis.setLightSourceType(lsType);		
	}
	public void setFullScreen(){
		dispose();
		this.setUndecorated(true);
		this.setResizable(false);
		Rectangle rect = GraphicsEnvironment.getLocalGraphicsEnvironment()
		.getMaximumWindowBounds();
		setSize(rect.width,rect.height);
		this.setLocation(0,0);
		setVisible(true);
		reflVis.repaint();
	}
	public void setMuseum(){
		this.getContentPane().remove(menuBar);
		List<BaseLightPattern<T>> lps = ctrl.getModel().getMuseumPatterns();
		reflVis.setTouchOverlay(new TouchButtonsOverlay(lps,menuBar));
		reflVis.repaint();
	}
}
