/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc.var;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;

import lattice.IBox;
import lattice.IShape;
import lattice.Ray;
import math.Util;
import math.VecMath;
import mvc.Controller;
import mvc.IModelListener;
import mvc.Model;
import mvc.ScanSession;
import mvc.scan.Observables;
import mvc.scan.ScanListener;
import mvc.scan.ScanVisualizer;

import common.Globals;

public class VariableVisualizer<T extends Number> extends JPanel {

	public static final int DIST_DIGITS = 4; 
	public final static String[] OBSERVABLES = 
			new String[]{"Reflections","Length","Reflections/Length"};
	private final short YSIDE_PIXEL = 270;
	private final short XSIDE_PIXEL = 200;
	private final Controller<T> ctrl;
	private JTextField obsHeightField;
	private JTextField obsDistField;
	private JTextField obsRotField;
	private JTextField obsSizeField;
	private JTextField obsDisplayField;
	private JTextField scRotMinField;
	private JTextField scRotMaxField;
	private JTextField scHeightMinField;
	private JTextField scHeightMaxField;
	private JTextField scDisplayField;
	private final JTextField[] allFields;
	private JTabbedPane tabPane;
	private JTextField msgField;
	private IModelListener modelListener;
	private ScanListener scanListener;
	private byte selObs = -1;
	private TitledBorder obsMaxTitle;
	private boolean truncate;
	private boolean showErrorMsg;
	public VariableVisualizer(final Controller<T> c) {
		this.ctrl = c;
		this.allFields = new JTextField[8];
		initGui();
		updateObservableFields(true);
		initListeners();
		msgField.setText("");
	}
	private void updateObservableFields(boolean force) {
		VecMath<T> math = ctrl.getModel().getMath();
		int[] carPos = new int[4];
		boolean hasFocus = false;
		JTextField[] tfs = new JTextField[]{obsHeightField,obsDistField,obsRotField,obsSizeField};
		for(int i = 0;i<tfs.length;i++) {
			if(tfs[i].hasFocus()) {
				JTextField tmp = tfs[i];
				tfs[i] = tfs[tfs.length-1];
				tfs[tfs.length-1] = tmp;
				hasFocus = true;
			}
			carPos[i] = tfs[i].getCaretPosition();
		}
		T y1 = ctrl.getModel().getRay().getStart()[1];
		if(truncate)
			y1 = math.util.round(y1,ctrl.getModel().getDecimalPlaces());
		boolean update = true;
		if(!force) {
			String txt = obsHeightField.getText();
			txt = txt.replace(BaseKeyListener.LIMIT_CHAR,' ');
			if(txt.trim().compareTo("")==0)
				update = true;
			else
				update = math.cmp(y1,math.util.parse(txt.trim()))!=0;
		}
		if(update) {
			String valTxt = math.util.toPlainString(y1);
			obsHeightField.setText(valTxt);
		}
		T x = ctrl.getModel().getRay().getStart()[0];
		if(truncate)
			x = math.util.round(x,DIST_DIGITS);
		if(!force) {
			String txt = obsDistField.getText();
			txt = txt.replace(BaseKeyListener.LIMIT_CHAR,' ');
			if(txt.trim().compareTo("")==0)
				update = true;
			else
				update = math.cmp(x,math.util.parse(txt.trim()))!=0;
		}
		if(update) {
			String valTxt = math.util.toPlainString(x);
			obsDistField.setText(valTxt);
		}
		T phi = ctrl.getModel().getGridRotation();
		phi = math.div(phi,math.PI);
		if(truncate)
			phi = math.util.round(phi,ctrl.getModel().getDecimalPlaces());
		if(!force) {
			String txt = obsRotField.getText();
			txt = txt.replace(BaseKeyListener.LIMIT_CHAR,' ');
			if(txt.trim().compareTo("")==0)
				update = true;
			else
				update = math.cmp(phi,math.util.parse(txt.trim()))!=0;
		}
		if(update) {
			String valTxt = math.util.toPlainString(phi);
			obsRotField.setText(valTxt);
		}
		T s = ctrl.getModel().getShape().getSize();
		if(truncate)
			s = math.util.round(s,ctrl.getModel().getDecimalPlaces());
		if(!force) {
			String txt = obsSizeField.getText();
			txt = txt.replace(BaseKeyListener.LIMIT_CHAR,' ');
			if(txt.trim().compareTo("")==0)
				update = true;
			else
				update = math.cmp(s,math.util.parse(txt.trim()))!=0;
		}
		if(update) {
			String valTxt = math.util.toPlainString(s);
			obsSizeField.setText(valTxt);
		}
		for(int i = 0;i<tfs.length;i++){
			int pos = Math.max(0,Math.min(carPos[i],tfs[i].getText().length()-1));
			tfs[i].setCaretPosition(pos);
		}
		if(!hasFocus)
			setMessage("");
		String txt;
		int rcnt = ctrl.getModel().getReflectionCount();
		if(rcnt>=0)
			txt = ""+rcnt;
		else
			txt = ">="+-rcnt;
		float length = ctrl.getModel().getLength();
		length = Math.round(length*100)/100f;
		if(length>=0)
			txt += "  /  "+length;
		else
			txt += "  /  >"+-length;
		try {
			int loops = ctrl.getModel().getLoops();
			if(loops>=0)
				txt += "  /  "+loops;
			else
				txt += "  /  >="+-loops;
		} catch(Exception e) {
			txt += "  /  error";
		}
		obsDisplayField.setText(txt);
	}
	private void initGui() {
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		setLayout(gridbag);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 3;
		c.gridheight = 1;
		tabPane = new JTabbedPane();
		tabPane.addTab("Main",createObservablesTab());
		tabPane.addTab("Scanner",createScannerTab());
		tabPane.setEnabledAt(1,false);
		gridbag.setConstraints(tabPane,c);
		add(tabPane);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 0.1;
		c.gridwidth = 1;
		msgField = new JTextField(" ");
		msgField.setEditable(false);
		msgField.setFocusable(false);
		msgField.setBorder(new EmptyBorder(1,1,1,1));
		msgField.setMinimumSize(new Dimension(100,35));
		Font f = msgField.getFont();
		msgField.setFont(new Font(f.getName(),f.getStyle(),f.getSize()-2));
		gridbag.setConstraints(msgField,c);
		add(msgField);
		c.insets = new Insets(10,5,5,0);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 1;
		c.gridy = 1;
		c.weightx = 0;
		c.weighty = 0.1;
		JTextField leftBtn = new JTextField("<");
		leftBtn.addMouseListener(new MouseAdapter() {

			@Override
			public void mousePressed(MouseEvent e) {
				for(JTextField tf:allFields)
					if(tf.hasFocus())
						tf.setCaretPosition(0);
			}
		});
		leftBtn.setEditable(false);
		leftBtn.setFocusable(false);
		leftBtn.setFont(new Font(Font.MONOSPACED,Font.BOLD,10));
		leftBtn.setBorder(new TitledBorder(""));
		leftBtn.setMinimumSize(new Dimension(20,0));
		gridbag.setConstraints(leftBtn,c);
		add(leftBtn);
		c.gridx = 2;
		JTextField rightBtn = new JTextField(">");
		rightBtn.setFocusable(false);
		rightBtn.addMouseListener(new MouseAdapter() {

			@Override
			public void mousePressed(MouseEvent e) {
				for(JTextField tf:allFields)
					if(tf.hasFocus())
						tf.setCaretPosition(tf.getText().length());
			}
		});
		rightBtn.setEditable(false);
		rightBtn.setFont(new Font(Font.MONOSPACED,Font.BOLD,10));
		rightBtn.setBorder(new TitledBorder(""));
		rightBtn.setMinimumSize(new Dimension(20,0));
		gridbag.setConstraints(rightBtn,c);
		add(rightBtn);
	}
	private JPanel createObservablesTab() {
		JPanel panel = new JPanel();
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		panel.setLayout(gridbag);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		obsHeightField = new JTextField();
		obsHeightField.setHorizontalAlignment(SwingConstants.RIGHT);
		obsHeightField.setCaretPosition(0);
		obsHeightField.addCaretListener(new PrintCaretPosListener() {

			@Override
			T getModelValue() {
				return ctrl.getModel().getRay().getStart()[1];
			}
		});
		JPanel p = wrapConf(obsHeightField,"Height");
		gridbag.setConstraints(p,c);
		panel.add(p);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		obsDistField = new JTextField();
		obsDistField.setHorizontalAlignment(SwingConstants.RIGHT);
		obsDistField.setCaretPosition(0);
		obsDistField.addCaretListener(new PrintCaretPosListener() {

			@Override
			T getModelValue() {
				return ctrl.getModel().getRay().getStart()[0];
			}
		});
		p = wrapConf(obsDistField,"Distance");
		gridbag.setConstraints(p,c);
		panel.add(p);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 2;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		obsRotField = new JTextField();
		obsRotField.setHorizontalAlignment(SwingConstants.RIGHT);
		obsRotField.setCaretPosition(0);
		obsRotField.addCaretListener(new PrintCaretPosListener() {

			@Override
			T getModelValue() {
				return ctrl.getModel().getGridRotation();
			}
		});
		p = createAnglePanel(obsRotField,"Angle");
		gridbag.setConstraints(p,c);
		panel.add(p);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 3;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		obsSizeField = new JTextField();
		obsSizeField.setHorizontalAlignment(SwingConstants.RIGHT);
		obsSizeField.setCaretPosition(0);
		obsSizeField.addCaretListener(new PrintCaretPosListener() {

			@Override
			T getModelValue() {
				return ctrl.getModel().getShape().getSize();
			}
		});
		p = wrapConf(obsSizeField,"Radius");
		gridbag.setConstraints(p,c);
		panel.add(p);
		c.fill = GridBagConstraints.BOTH;
		c.insets = new Insets(0,0,0,0);
		c.gridx = 0;
		c.gridy = 4;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		String obsName = OBSERVABLES[Observables.OBS_REFL_CNT];
		obsName += " / "+OBSERVABLES[Observables.OBS_LENGTH];
		obsName += " / Loops";
		obsDisplayField = new JTextField();
		obsDisplayField.setHorizontalAlignment(SwingConstants.RIGHT);
		obsDisplayField.setEditable(false);
		obsDisplayField.setFocusable(false);
		obsDisplayField.setBorder(new TitledBorder(obsName));
		obsDisplayField.setForeground(Color.gray.darker());
		//Color col = obsDisplayField.getBackground();
		//obsDisplayField.setBackground(Globals.makeDarker(col));
		gridbag.setConstraints(obsDisplayField,c);
		panel.add(obsDisplayField);
		allFields[0] = obsHeightField;
		allFields[1] = obsDistField;
		allFields[2] = obsRotField;
		allFields[3] = obsSizeField;
		panel.setVisible(true);
		return panel;
	}
	private JPanel createScannerTab() {
		JPanel panel = new JPanel();
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		panel.setLayout(gridbag);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		scRotMinField = new JTextField();
		scRotMinField.setHorizontalAlignment(SwingConstants.RIGHT);
		scRotMinField.addCaretListener(new PrintCaretPosListener() {

			@Override
			T getModelValue() {
				IBox<T> frame = ctrl.getModel().getScanSession().getInitBox();
				return frame.getCorner()[0];
			}
		});
		JPanel p = createAnglePanel(scRotMinField,"Min Angle");
		gridbag.setConstraints(p,c);
		panel.add(p);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		scRotMaxField = new JTextField();
		scRotMaxField.setHorizontalAlignment(SwingConstants.RIGHT);
		scRotMaxField.addCaretListener(new PrintCaretPosListener() {

			@Override
			T getModelValue() {
				IBox<T> frame = ctrl.getModel().getScanSession().getInitBox();
				VecMath<T> math = ctrl.getModel().getMath();
				return math.add(frame.getCorner()[0],frame.getLengths()[0]);
			}
		});
		p = createAnglePanel(scRotMaxField,"Max Angle");
		gridbag.setConstraints(p,c);
		panel.add(p);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 2;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		scHeightMinField = new JTextField();
		scHeightMinField.setHorizontalAlignment(SwingConstants.RIGHT);
		scHeightMinField.addCaretListener(new PrintCaretPosListener() {

			@Override
			T getModelValue() {
				IBox<T> frame = ctrl.getModel().getScanSession().getInitBox();
				return frame.getCorner()[1];
			}
		});
		p = wrapConf(scHeightMinField,"Min Height");
		gridbag.setConstraints(p,c);
		panel.add(p);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 3;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		scHeightMaxField = new JTextField();
		scHeightMaxField.setHorizontalAlignment(SwingConstants.RIGHT);
		scHeightMaxField.addCaretListener(new PrintCaretPosListener() {

			@Override
			T getModelValue() {
				IBox<T> frame = ctrl.getModel().getScanSession().getInitBox();
				VecMath<T> math = ctrl.getModel().getMath();
				return math.add(frame.getCorner()[1],frame.getLengths()[1]);
			}
		});
		p = wrapConf(scHeightMaxField,"Max Height");
		gridbag.setConstraints(p,c);
		panel.add(p);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 4;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 1;
		scDisplayField = new JTextField();
		scDisplayField.setHorizontalAlignment(SwingConstants.RIGHT);
		scDisplayField.setEditable(false);
		obsMaxTitle = new TitledBorder("Observable Max");
		scDisplayField.setBorder(obsMaxTitle);
		//Color col = scDisplayField.getBackground();
		//scDisplayField.setBackground(Globals.makeDarker(col));
		gridbag.setConstraints(scDisplayField,c);
		panel.add(scDisplayField);
		allFields[4] = scHeightMinField;
		allFields[5] = scHeightMaxField;
		allFields[6] = scRotMinField;
		allFields[7] = scRotMaxField;
		panel.setVisible(true);
		return panel;
	}
	private static JPanel createAnglePanel(JTextField tf,String title) {
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		JPanel p = new JPanel(gbl);
		p.setBorder(new TitledBorder(title));
		gbc.fill = GridBagConstraints.BOTH;
		gbc.gridx = 0;
		gbc.gridy = 0;
		gbc.weightx = 1;
		gbc.weighty = 1;
		p.setBackground(tf.getBackground());
		tf.setBackground(Globals.INPUT_BG);
		gbl.setConstraints(tf,gbc);
		p.add(tf);
		gbc.fill = GridBagConstraints.BOTH;
		gbc.gridx = 1;
		gbc.gridy = 0;
		gbc.weightx = 0;
		gbc.weighty = 1;
		gbc.insets = new Insets(0,5,0,0);
		JTextField piLabel = new JTextField(" * \u03C0"/*pi*/);
		piLabel.setEditable(false);
		piLabel.setBackground(Globals.INPUT_BG);
		piLabel.setMinimumSize(piLabel.getPreferredSize());
		gbl.setConstraints(piLabel,gbc);
		p.add(piLabel);
		return p;
	}
	private JPanel wrapConf(JTextField tf,String borderText){
		tf.setBackground(Globals.INPUT_BG);
		JPanel l = new JPanel(new GridLayout());
		l.add(tf);
		TitledBorder tb = BorderFactory.createTitledBorder(borderText);
		l.setBorder(tb);
		return l;
	}
	public void updateScannerFrameFields(IBox<T> frame) {
		VecMath<T> math = ctrl.getModel().getMath();
		int rndPos = Integer.MIN_VALUE;
		if(truncate) {
			T[] lengths = frame.getLengths().clone();
			lengths[0] = math.div(lengths[0],math.util.cast(1000));
			lengths[1] = math.div(lengths[1],math.util.cast(1000));
			int s1 = math.util.getScale(lengths[0]);
			int s2 = math.util.getScale(lengths[1]);
			rndPos = Math.max(-s1,-s2);
		}
		T phi1 = frame.getCorner()[0];
		phi1 = math.div(phi1,math.PI);
		if(truncate)
			phi1 = math.util.round(phi1,rndPos);
		String s = math.util.toPlainString(phi1);
		scRotMinField.setText(s);
		scRotMinField.setCaretPosition(0);
		T phi2 = math.add(frame.getCorner()[0],frame.getLengths()[0]);
		phi2 = math.div(phi2,math.PI);
		if(truncate)
			phi2 = math.util.round(phi2,rndPos);
		s = math.util.toPlainString(phi2);
		scRotMaxField.setText(s);
		scRotMaxField.setCaretPosition(0);
		T h1 = frame.getCorner()[1];
		if(truncate)
			h1 = math.util.round(h1,rndPos);
		s = math.util.toPlainString(h1);
		scHeightMinField.setText(s);
		scHeightMinField.setCaretPosition(0);
		T h2 = math.add(frame.getCorner()[1],frame.getLengths()[1]);
		if(truncate)
			h2 = math.util.round(h2,rndPos);
		s = math.util.toPlainString(h2);
		scHeightMaxField.setText(s);
		scHeightMaxField.setCaretPosition(0);
		if(selObs==-1) {
			ScanVisualizer<T> scanVis = ctrl.getView().getScanVisualizer();
			if(scanVis!=null)
				setObservable(scanVis.getObservable(),0);
		}
		setScannerTabEnabled(true);
	}
	public void updateScannerObsFields(Observables obs) {
		String txt = obs.getMax(selObs)+"  /  "+obs.aproxPixels;
		scDisplayField.setText(txt);
		//scPixelField.setText(obs.aproxPixels+"");
	}
	private void initListeners() {
		modelListener = new IModelListener() {

			@Override
			public void modelChanged(byte event) {
				switch(event) {
					case Model.RAY_NEW:
						if(obsDisplayField.getText().compareTo("")!=0) {
							//obsReflCountField.setText("");
							obsDisplayField.setText("");
						}
						Cursor wcur = Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR);
						if(getCursor()!=wcur) {
							setCursor(wcur);
							for(int i = 0;i<4;i++)
								allFields[i].setCursor(wcur);
						}
						break;
					case Model.RAY_FINISHED:
						setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						Cursor tcur = Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR);
						for(int i = 0;i<4;i++)
							allFields[i].setCursor(tcur);
						updateObservableFields(false);
						break;
					//case Model.MODEL_WORLD_RESIZE:
					//	updateObservableFields(false);
				}
			}
		};
		this.ctrl.getModel().addModelListener(modelListener);
		scanListener = new ScanListener() {

			boolean init = false;
			@Override
			public void start() {
				init();
			}
			@Override
			public void stop(String msg) {}
			@Override
			public void update(int x,int y) {
				if(!init)
					init();
				ScanSession<T> scs = ctrl.getModel().getScanSession();
				updateScannerObsFields(scs.obs);
			}
			private void init() {
				ScanSession<T> scs = ctrl.getModel().getScanSession();
				updateScannerFrameFields(scs.getInitBox());
				init = true;
			}
		};
		this.ctrl.getModel().addScanListener(scanListener);
		obsHeightField.addKeyListener(new RayKeyListener<T>(ctrl,obsHeightField,false) {

			@Override
			void handleValue(T val) {
				Ray<T> ray = ctrl.getModel().getRay();
				ray.getStart()[1] = val;
				ctrl.getModel().setRay(ray,Model.STATIC_CALC);
			}
		});
		obsDistField.addKeyListener(new RayKeyListener<T>(ctrl,obsDistField,true) {

			@Override
			void handleValue(T val) {
				Ray<T> ray = ctrl.getModel().getRay();
				ray.getStart()[0] = val;
				ctrl.getModel().setRay(ray,Model.STATIC_CALC);
			}
		});
		obsRotField.addKeyListener(new RayKeyListener<T>(ctrl,obsRotField,true) {

			@Override
			void handleValue(T val) {
				Model<T> model = ctrl.getModel();
				T phi = model.getMath().mul(val,model.getMath().PI);
				model.setGridRotation(phi);
			}
		});
		obsSizeField.addKeyListener(new RayKeyListener<T>(ctrl,obsSizeField,true) {

			@Override
			void handleValue(T r) {
				Model<T> model = ctrl.getModel();
				T MAX_R = model.transformShapeSize((byte)100);
				VecMath<T> mt = model.getMath();
				IShape<T> s = model.getShape();
				if(mt.cmp(r,MAX_R)<0)
					s.setSize(r);
				else
					s.setSize(MAX_R);
				model.setShape(s);
			}
		});
		KeyListener gskl = new GroupedScanKeyListener<T>(ctrl,scRotMinField,scRotMaxField,
		        scHeightMinField,scHeightMaxField);
		scRotMinField.addKeyListener(gskl);
		scRotMaxField.addKeyListener(gskl);
		scHeightMinField.addKeyListener(gskl);
		scHeightMaxField.addKeyListener(gskl);
	}
	@Override
	public Dimension getPreferredSize() {
		return new Dimension(XSIDE_PIXEL,YSIDE_PIXEL);
	}
	public void setScannerTabEnabled(boolean b) {
		tabPane.setEnabledAt(1,b);
		if(!b)
			tabPane.setSelectedIndex(0);
	}
	public void onDelete() {
		ctrl.getModel().removeScanListener(scanListener);
		ctrl.getModel().removeModelListener(modelListener);
	}
	public void setObservable(byte idx,Number max) {
		scDisplayField.setText(max.toString());
		selObs = idx;
		String obsName = OBSERVABLES[selObs];
		obsMaxTitle.setTitle("Max. "+obsName+" / Prec. Errors");
	}
	public synchronized void enableTruncation(final boolean b) {
		SwingUtilities.invokeLater(new Runnable() {

			public void run() {
				truncate = b;
				updateObservableFields(true);
				if(tabPane.isEnabledAt(1)) {
					ScanSession<T> scs = ctrl.getModel().getScanSession();
					if(scs!=null)
						updateScannerFrameFields(scs.getInitBox());
				}
			}
		});
	}
	void setMessage(String msg) {
		msgField.setText(msg);
	}
	abstract class PrintCaretPosListener implements CaretListener {

		@Override
		public void caretUpdate(CaretEvent e) {
			JTextField src = (JTextField)e.getSource();
			String txt = src.getText().trim();
			int pos = txt.indexOf(".");
			int digits = txt.length()-(pos!=-1?1:0);
			if(e.getDot()>-1) {
				int dpos1 = e.getDot();
				int dpos2 = pos!=-1&&pos<=dpos1?dpos1:dpos1+1;
				String msg = "Digit: "+Math.min(dpos2,digits)+"/"+digits;
				if(showErrorMsg){
					Util<T> util = ctrl.getModel().getMath().util;
					int errScale = util.getErrorScale(getModelValue());
					if(errScale==Util.MIN_INT)
						msg += ", Error=0";
					else
						msg += ", Error<10e"+(errScale+1);
				}
				msgField.setText(msg);
			}
		}
		abstract T getModelValue();
	}
	public void enableErrorMsg(final boolean b) {
		SwingUtilities.invokeLater(new Runnable() {

			public void run() {
				showErrorMsg = b;
				updateObservableFields(true);
				if(tabPane.isEnabledAt(1)) {
					ScanSession<T> scs = ctrl.getModel().getScanSession();
					if(scs!=null)
						updateScannerFrameFields(scs.getInitBox());
				}
			}
		});
	}
}
