/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package parser;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.StringTokenizer;

public final class ExpressionTree<T extends Number> {

	private final Expression<T> root;
	public ExpressionTree(Expression<T> root) {
		this.root = root;
	}
	public T calculate(HashMap<String,T> map) throws Exception {
		Collection<Expression<T>> exprs = getExpressions();
		for(Expression<T> expr:exprs)
			if(expr instanceof ValueExpression) {
				if(!insert(map,(ValueExpression<T>)expr)) {
					String var = ((ValueExpression<T>)expr).getVariable();
					throw new Exception("Variable "+var+" has no value.");
				}
			}
		return root.getValue();
	}
	public T calculate(String args,T[] values) throws Exception {
		return calculate(buildValueMap(args,values));
	}
	private boolean insert(HashMap<String,T> varVals,ValueExpression<T> value) {
		boolean ok = false;
		if(value.getVariable()!=null) {
			if(value.getVariable().compareTo(Symbols.EULER)==0) {
				value.setValue(root.math.util.getE(root.math.PRECISION+100));
				ok = true;
			} else if(value.getVariable().compareTo(Symbols.PI)==0) {
				value.setValue(root.math.util.getPi(root.math.PRECISION+100));
				ok = true;
			} else
				for(String var:varVals.keySet())
					if(value.getVariable().compareTo(var)==0) {
						value.setValue(varVals.get(var));
						ok = true;
						break;
					}
		} else
			ok = true;
		return ok;
	}
	public Expression<T> getRoot() {
		return root;
	}
	public Collection<Expression<T>> getExpressions() {
		ArrayList<Expression<T>> nodeList = new ArrayList<Expression<T>>();
		addNodes(root,nodeList);
		return nodeList;
	}
	private void addNodes(Expression<T> expr,Collection<Expression<T>> exprList) {
		exprList.add(expr);
		for(int i = 0;i<expr.getChildCount();i++)
			addNodes(expr.getChild(i),exprList);
	}
	public HashMap<String,T> buildValueMap(String args,T[] values) throws Exception {
		HashMap<String,T> retVal = new HashMap<String,T>();
		StringTokenizer st = null;
		if(args!=null)
			st = new StringTokenizer(args.trim(),Symbols.DELIMITER);
		while(st!=null&&st.hasMoreTokens()) {
			String paramDef = st.nextToken();
			int idx = paramDef.indexOf("=");
			if(idx!=-1) {
				String key = paramDef.substring(0,idx).trim();
				String val = paramDef.substring(idx+1).trim();
				if(key==null||key.compareTo("")==0)
					throw new Exception("Mising parameter: "+paramDef);
				if(val==null||val.compareTo("")==0)
					throw new Exception("Mising value: "+paramDef);
				int pr = root.math.PRECISION+100;
				T value;
				if(val.indexOf(Symbols.REFERENCE)==0) {
					int row;
					try {
						row = Integer.parseInt(val.substring(1))-1;
					} catch(Exception e) {
						throw new Exception("$ must be followed by an integer");
					}
					if(row>=values.length)
						throw new Exception("Value reference points to an subsequent row");
					value = values[row];
					if(value==null)
						throw new Exception("Value reference points to an empty row");
					if(root.math.util.precision(value)>pr)
						value = root.math.util.copy(value,pr);
				} else if(val.compareTo(Symbols.EULER)==0)
					value = root.math.util.getE(root.math.PRECISION+100);
				else if(val.compareTo(Symbols.PI)==0)
					value = root.math.util.getPi(root.math.PRECISION+100);
				else
					try {
						value = root.math.util.parse(val);
						value = root.math.util.copy(value,root.math.PRECISION+100);
					} catch(Exception e) {
						throw new Exception("Can't parse number: "+paramDef);
					}
				retVal.put(key,value);
			} else
				throw new Exception("Missing = in parameter list: "+paramDef);
		}
		return retVal;
	}
	public boolean containsVariable(String var) {
		Collection<Expression<T>> exprs = getExpressions();
		for(Expression<T> expr:exprs)
			if(expr instanceof ValueExpression) {
				ValueExpression<T> valExpr = (ValueExpression<T>)expr;
				String v = valExpr.getVariable();
				if(v!=null&&v.compareTo(var)==0)
					return true;
			}
		return false;
	}
}
